#! /usr/bin/env python3
#
# Test(s) for ../summarize_pytest.py
#
# Notes:
# - This can be run as follows (e.g., from root of repo):
#   $ pytest ./mezcla/examples/tests/test_summarize_pytest.py
#
#...............................................................................
# Sample tested input:
#
#   =================== 1 passed, 3 skipped, 3 xpassed in 1.23s ====================
#   ============================== 2 xpassed in 0.73s ==============================
#   ======================= 1 xfailed, 7 warnings in 36.99s ========================
#   ============================== 2 skipped in 0.24s ==============================
#   ======================== 1 xfailed, 1 xpassed in 1.64s =========================
#
# Alternative input:
#  tests/test_example.09jan25.12.out:================================= 22 failed, 44 passed, 4 skipped in 2.19s =================================
#  tests/test_another.09jan25.13.out:================================= 10 passed, 2 xfailed, 1 xpassed in 1.50s =================================
#  tests/test_third.09jan25.14.out:================================= 5 failed, 15 passed in 0.75s =================================
#

"""Tests for summarize_pytest module"""

# Standard modules
## TODO: from typing import Optional

# Installed modules
import pytest

# Local modules
from mezcla import debug
from mezcla.my_regex import my_re
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests

# Note: Two references are used for the module to be tested:
#    THE_MODULE:               module object (e.g., <module 'mezcla.main' ...>)
#    TestIt.script_module:     dotted module path (e.g., "mezcla.main")
THE_MODULE = None
try:
    import mezcla.summarize_pytest as THE_MODULE
except:
    system.print_exception_info("summarize_pytest import") 

#------------------------------------------------------------------------

class TestIt(TestWrapper):
    """Class for command-line based testcase definition"""
    # note: script_module used in argument parsing sanity check (e.g., --help)
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    def setUp(self):
        """Per-test setup"""
        #
        debug.trace(6, f"TestIt.setUp(); self={self}")
        # note: must do parent processing first (e.g., for temp file support)
        super().setUp()
        
        # Create sample pytest output lines for testing
        self.sample_lines = [
            "tests/test_example.09jan25.12.out:================================= 22 failed, 44 passed, 4 skipped in 2.19s =================================",
            "tests/test_another.09jan25.13.out:================================= 10 passed, 2 xfailed, 1 xpassed in 1.50s =================================",
            "tests/test_third.09jan25.14.out:================================= 5 failed, 15 passed in 0.75s ================================="
        ]
        # TODO: debug.trace_current_context(level=debug.QUITE_DETAILED)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_01_data_file(self):
        """Tests run_script w/ data file"""
        # Warning: see notes above about potential issues with run_script-based tests.
        debug.trace(4, f"TestIt.test_01_data_file(); self={self}")
        system.write_lines(self.temp_file, self.sample_lines)
        output = self.run_script(options="", env_options="",
                                 data_file=self.temp_file)
        self.do_assert(my_re.search(r"Timestamp.*Pass.*Fail", output.strip()))
        self.do_assert("09jan25.12" in output)
        self.do_assert("44" in output)  # passed count
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_02_parse_single_line(self):
        """Test for parsing a single pytest output line"""
        debug.trace(4, f"TestIt.test_02_parse_single_line(); self={self}")
        summarizer = THE_MODULE.PytestSummarizer()
        result = summarizer.parse_pytest_line(self.sample_lines[0])
        
        self.do_assert(result is not None)
        self.do_assert(result.passed == 44)
        self.do_assert(result.failed == 22)
        self.do_assert(result.skipped == 4)
        self.do_assert(abs(result.duration - 2.19) < 0.01)
        self.do_assert(result.timestamp == '09jan25.12')
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_03_process_multiple_lines(self):
        """Test processing multiple pytest output lines"""
        debug.trace(4, f"TestIt.test_03_process_multiple_lines(); self={self}")
        summarizer = THE_MODULE.PytestSummarizer()
        results = summarizer.process(self.sample_lines)
        
        self.do_assert(len(results) == 3)
        self.do_assert(results[0].timestamp == '09jan25.12')
        self.do_assert(results[1].timestamp == '09jan25.13')
        self.do_assert(results[2].timestamp == '09jan25.14')
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_04_test_result_calculations(self):
        """Test TestResult calculations for total tests and OK rate"""
        debug.trace(4, f"TestIt.test_04_test_result_calculations(); self={self}")
        
        result = THE_MODULE.TestResult(
            passed=10, failed=5, xfailed=2, xpassed=1
        )
        
        # Test without treating xresults
        self.do_assert(result.total_tests(treat_xresults=False) == 15)  # 10 + 5
        self.do_assert(abs(result.ok_rate(treat_xresults=False) - 66.67) < 0.1)
        
        # Test with treating xresults
        self.do_assert(result.total_tests(treat_xresults=True) == 18)  # 10 + 5 + 2 + 1
        expected_ok = (10 + 1) * 100.0 / 18  # passed + xpassed
        self.do_assert(abs(result.ok_rate(treat_xresults=True) - expected_ok) < 0.1)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_05_ignore_timestamp(self):
        """Test ignoring timestamp requirement"""
        debug.trace(4, f"TestIt.test_05_ignore_timestamp(); self={self}")
        
        line_without_timestamp = "================================= 10 passed in 1.0s ================================="
        summarizer = THE_MODULE.PytestSummarizer(ignore_timestamp=True)
        result = summarizer.parse_pytest_line(line_without_timestamp)
        self.do_assert(result is not None)
        self.do_assert(result.timestamp == 'n/a')
        self.do_assert(result.passed == 10)
        self.do_assert(result.total_tests() == 10)
        self.do_assert(abs(result.duration - 1.0) < 0.01)

        line_without_timestamp = "=== 1 failed, 5 passed, 2 skipped, 17 xfailed, 39 xpassed, 13 warnings in 139.11s (0:02:19) ==="
        summarizer = THE_MODULE.PytestSummarizer(ignore_timestamp=True)
        result = summarizer.parse_pytest_line(line_without_timestamp)
        self.do_assert(result is not None)
        self.do_assert(result.timestamp == 'n/a')
        self.do_assert(result.failed == 1)
        self.do_assert(result.passed == 5)
        self.do_assert(result.skipped == 2)
        self.do_assert(result.xfailed == 17)
        self.do_assert(result.xpassed == 39)
        self.do_assert(result.total_tests(treat_xresults=False) == 6)
        self.do_assert(result.total_tests(treat_xresults=True) == 62)
        self.do_assert(abs(result.duration - 139) < 0.5)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_06_combine_results(self):
        """Test combining results with same timestamp"""
        debug.trace(4, f"TestIt.test_06_combine_results(); self={self}")
        
        lines_same_timestamp = [
            "test1.09jan25.out:================================= 10 passed in 1.0s =================================",
            "test2.09jan25.out:================================= 5 passed, 2 failed in 0.5s ================================="
        ]
        
        summarizer = THE_MODULE.PytestSummarizer(combine=True)
        results = summarizer.process(lines_same_timestamp)
        
        self.do_assert(len(results) == 1)
        self.do_assert(results[0].passed == 15)  # 10 + 5
        self.do_assert(results[0].failed == 2)
        self.do_assert(abs(results[0].duration - 1.5) < 0.01)  # 1.0 + 0.5
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_07_format_output(self):
        """Test formatting output as string table"""
        debug.trace(4, f"TestIt.test_07_format_output(); self={self}")
        
        summarizer = THE_MODULE.PytestSummarizer()
        results = summarizer.process(self.sample_lines[:1])
        output = summarizer.format_output(results)
        
        self.do_assert("Timestamp" in output)
        self.do_assert("Pass" in output)
        self.do_assert("OK-pct" in output)
        self.do_assert("09jan25.12" in output)
        self.do_assert("44" in output)  # passed count
        self.do_assert("22" in output)  # failed count
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_08_cli_header_only(self):
        """Test CLI with header-only option"""
        debug.trace(4, f"TestIt.test_08_cli_header_only(); self={self}")
        
        system.write_lines(self.temp_file, self.sample_lines)
        output = self.run_script(
            options=f"--{THE_MODULE.HEADER_ONLY_OPT}",
            data_file=self.temp_file
        )
        
        # Should have header but no data
        self.do_assert("Timestamp" in output)
        self.do_assert("09jan25" not in output)  # No data rows
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_09_cli_treat_xresults(self):
        """Test CLI with treat-xresults option"""
        debug.trace(4, f"TestIt.test_09_cli_treat_xresults(); self={self}")
        
        # Use line with xfailed and xpassed
        system.write_lines(self.temp_file, self.sample_lines)
        output = self.run_script(
            options=f"--{THE_MODULE.TREAT_XRESULTS_OPT}",
            data_file=self.temp_file
        )
        
        self.do_assert("09jan25.13" in output)
        # With treat_xresults, total should be 13 (10 + 2 + 1)
        self.do_assert(my_re.search(r"\s+13\s+", output))
        return

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python
#
# Customization of PyLucene SearchFiles.py for searching index of tabular file:
#    http://svn.apache.org/viewvc/lucene/pylucene/trunk/samples/SearchFiles.py
# This also includes an option to do similar searching via pysolr.
#
# Notes:
# - Searches the field 'contents' by default (as indexed by index_table_file.py).
# - The text returned by PyLucene routines can be in Unicode, so it should be converted
#   into UTF-8 under Python 2.x (see comments in show_unicode_info.py for details).
# - The Solr implementation is not as comprehensive as the Lucene support (e.g., term vectors not yet supported).
#
# Warning:
# - This is basically the same version from the misc-scripts repo,
#   so the style is archaic compared to other mezcla scripts.
# - This is mainly included for sake of completeness, because
#   analyze_tfidf.py got added to the repo inadvertantly.
# - It is also used in the test_format_profile.py as unlikely to change.
#
# TODO:
# - *** Add support for searching ngram tokens ***
# - Have option to use relative term frequency: tf(t, d) = f(t, d) / max(f(t', d) for t' in d).
# - Add TF/IDF support to interactive queries (e.g., via run()).
# - Track down discrepancy in document numbers in TF/IDF listing vs. query search.
# - Convert getenv-based arguments to argparse.
# - * Add sanity check for PyLucene 4.4 *
# - Add optional smoothing for TF/IDF.
# - Use remote_dispatching.py to encapsulate more of the remote processing support.
#

"""
Usage: {script} [index_dir [query_word1 ...]]

This script is loosely based on the Lucene demo sample SearchFiles.java
(i.e., org.apache.lucene.demo.SearchFiles).  It prompts for a search query, 
then it searches the Lucene index in the current directory called 'index' for the
search query entered against the 'contents' field.  It then displays the
'name' and 'contents' fields for each of the hits it finds in the index.  Note that
search.close() is currently commented out because it causes a stack overflow in
some cases.

Examples:

$ {script}  table-file-index  programmer  cashier

$ echo $'truck driver\\nsales associate\\n' | PHRASAL=1 INDEX_DIR=table-file-index {script} -

Notes:
- Can search against a Solr server, using 'id' and 'content' fields.
- The processing can be customized via the following environment variables:
	{descriptions}
- Use - for the query to search all documents (or all terms for DOC_FREQ).
"""

import logging
import math
import os
import re
import sys

from mezcla import glue_helpers as gh
from mezcla import tpo_common as tpo
from mezcla.tpo_common import getenv_text, getenv_boolean, getenv_integer, debug_print, print_stderr
from mezcla import system

USE_SOLR = getenv_boolean("USE_SOLR", False, "Query Solr server instead of Lucene")
USE_LUCENE = getenv_boolean("USE_LUCENE", not USE_SOLR, "Query Solr server instead of Lucene")
if USE_SOLR:
    import pysolr
elif USE_LUCENE:
    import lucene
    from java.io import File
    from org.apache.lucene.analysis.standard import StandardAnalyzer

    from org.apache.lucene.index import DirectoryReader, MultiFields, Term, TermsEnum
    from org.apache.lucene.queryparser.classic import QueryParser
    from org.apache.lucene.store import SimpleFSDirectory
    from org.apache.lucene.search import IndexSearcher, Query
    from org.apache.lucene.util import Version
    from org.apache.lucene.util import BytesRefIterator
else:
    debug_print("Warning: requires either USE_SOLR and USE_LUCENE", 4)
    pass

#------------------------------------------------------------------------
# Globals

INDEX_DIR = getenv_text("INDEX_DIR", "table-file-index", "Directory where Lucene index resides")
# Note: Solr by defaul uses id & content whereas PyLucene demo used name & contents for ID and data fields
FIELD = getenv_text("FIELD", ("content" if USE_SOLR else "contents"), "Field name for searching")
NAME = getenv_text("NAME", ("id" if USE_SOLR else "name"), "Name/ID field for results")
JUST_HITS = getenv_boolean("JUST_HITS", False, "Just show document hit count")
MAX_HITS = int(getenv_text("MAX_HITS", 50, "Maximum number of hits to return"))
SHOW_FREQ = getenv_boolean("SHOW_FREQ", False, "Show document frequency for terms")
PHRASAL = getenv_boolean("PHRASAL", False, "Use phrasal search")
TOKEN_REGEX = getenv_text("TOKEN_REGEX", ("\n" if PHRASAL else r"\W+"), "Regex for tokenizing input")
VERBOSE = getenv_boolean("VERBOSE", False, "Verbose output mode")
COUNT_TERMS = getenv_boolean("COUNT_TERMS", False, "Count number of terms in entire index")
SHOW_TFIDF = getenv_boolean("SHOW_TFIDF", False, "Show TF/IDF-based differntiating terms")
SKIP_UNIGRAMS = getenv_boolean("SKIP_UNIGRAMS", False,
                               "Don't include unigrams in TF/IDF listing")
SKIP_PLACEHOLDERS = getenv_boolean("SKIP_PLACEHOLDERS", False,
                                   "Skip phrasal terms with placeholders (e.g., '_ california')")
MAX_TERMS = getenv_integer("MAX_TERMS", 10, "Maximum number of terms for TF/IDF")
USE_NAME_FOR_ID = getenv_boolean("USE_NAME_FOR_ID", False, "Use the name field in place of Lucene docid")
REVERSE_ORDER = getenv_boolean("REVERSE_ORDER", False, "Produce document listing in reverse")
RETAIN_DOC_ID = getenv_boolean("RETAIN_DOC_ID", False, "Use Lucene docid in output")
SOLR_URL = getenv_text("SOLR_URL", "http://localhost:8983/solr", "URL for Solr server to query")
USE_DISMAX = getenv_boolean("USE_DISMAX", False, "Use DISMAX with Solr")
CONJUNCTIVE = getenv_boolean("CONJUNCTIVE", False, "Use conjunctive query with Solr")
#
DISTRIBUTED = getenv_boolean("DISTRIBUTED", False, "Run distributed with input partitioned and entire index copied")
## REMOTE_WORKERS = getenv_text("REMOTE_WORKERS", "", "String list of workers for remote distribution")
START_DOC = getenv_integer("START_DOC", 0)
SLEEP_SECONDS = getenv_integer("SLEEP_SECONDS", 60)
MAX_SLEEP = getenv_integer("MAX_SLEEP", 86400)
SKIP_INDEX = getenv_boolean("SKIP_INDEX", False)
## TODO: NUM_DOCS = getenv_integer("NUM_DOCS", system.MAX_INT)
## TODO: END_DOC = getenv_integer("END_DOC", START_DOC + NUM_DOCS)
START_DOC = getenv_integer("START_DOC", system.MAX_INT)
# Debugging options
MAX_EMPTY = getenv_integer("MAX_EMPTY", 1000)
IGNORE_EXCEPTIONS = getenv_boolean("IGNORE_EXCEPTIONS", (tpo.debugging_level() >= tpo.LEVEL4))


#------------------------------------------------------------------------

class IndexLookup(object):
    """Class for performing queries against a Lucene index"""
    vm_obj = None

    def __init__(self, index_dir=INDEX_DIR, field=FIELD):
        debug_print("IndexLookup.__init__(%s, %s)" % (index_dir, field), 5)
        self.field = field
        self.frequencies = {}
        if USE_SOLR:
            self.init_solr()
        else:
            self.init_lucene(index_dir)
        return

    def init_solr(self):
        """Initialize Solr processing"""
        debug_print("IndexLookup.init_solr()", 5)
        #
        # Enable detailed Solr tracing if debugging
        if (tpo.verbose_debugging()):
            ## solr_logger = logging.getLogger('pysolr')
            ## solr_logger.setLevel(logging.DEBUG)
            # TODO: use mapping from symbolic LEVEL user option (e.g., via getenv)
            level = logging.INFO if (tpo.debug_level < 4) else logging.DEBUG
            logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=level)
        # format: Solr(url, [decoder=None], [timeout=60])
        # TODO: add option for timeout
        self.solr = pysolr.Solr(SOLR_URL)
        tpo.debug_print("solr=%s" % self.solr, 4)
        return

    def init_lucene(self, index_dir):
        """Initialize Lucene processing"""
        debug_print("IndexLookup.init_lucene(%s)" % index_dir, 5)
        if not IndexLookup.vm_obj:
            try:
                # Issue standard pylucene startup sequence
                # Note: headless is for vm's without graphical display
                IndexLookup.vm_obj = lucene.initVM(vmargs=['-Djava.awt.headless=true'])
                tpo.debug_format("vm_obj={vm}", 5, vm=IndexLookup.vm_obj)
            except:
                debug_print("Warning: Exception during IndexLookup __init__: %s" % str(sys.exc_info()), 4)
        debug_print('lucene version: %s' % lucene.VERSION, 3)
        directory = SimpleFSDirectory(File(index_dir))
        self.reader = DirectoryReader.open(directory)
        self.searcher = IndexSearcher(self.reader)
        self.analyzer = StandardAnalyzer(Version.LUCENE_CURRENT)
        return

    def run(self):
        """Interactive run queries specified on the console"""
        debug_print("IndexLookup.run()", 5)
        query_prompt = ("Query:" if VERBOSE else "")
        while True:
            if (VERBOSE):
                print("")
                print("Hit enter with no input to quit.")
            try:
                command = raw_input(query_prompt)
            except:
                break
            if command.strip() == '':
                break

            if SHOW_FREQ:
                gh.assertion(not USE_SOLR)
                tokens = [t for t in re.split(TOKEN_REGEX, command) if t]
                for term in tokens:
                    print("%s: %d" % (term, self.doc_freq(term)))
            elif JUST_HITS:
                print(self.get_hit_count(command))
            else:
                doc_names = self.issue_query(command)
                print("\n".join(doc_names))
        return

    def get_num_docs(self):
        """Returns number of documents in the Lucene index"""
        gh.assertion(not USE_SOLR)
        num_docs = 0
        try:
            num_docs = self.reader.numDocs()
        except:
            if (not IGNORE_EXCEPTIONS):
                tpo.debug_raise()
            debug_print("Warning: Exception during get_num_docs: %s" % str(sys.exc_info()), 4)
        debug_print("get_num_docs() => %d" % num_docs, 4)
        return (num_docs)

    def get_doc_terms_and_freqs(self, doc_id):
        """Return tuple with list of all indexed terms and list of their frequencies for DOC_ID"""
        gh.assertion(not USE_SOLR)
        if not isinstance(doc_id, int):
            doc_id = int(doc_id)
        terms = []
        tfs = []
        try:
            tv = self.reader.getTermVector(doc_id, self.field)
            if not tv:
                # Note: To distinguish between field with no entries and field w/o term vectors,
                # the document contents are retrieved. For this to work, the field must be stored
                # (e.g., via STORE_CONTENTS setting of index_table_file.py).
                # via http://lucene.apache.org/core/4_0_0/core/org/apache/lucene/document/Document.html
                #     Fields which are not stored are not available in documents retrieved from the index.
                # TODO: work out a way to detect this without storing the field
                doc = self.reader.document(doc_id)
                contents = doc.get(self.field) if doc else ""
                if contents:
                    print_stderr("Error: No term vectors for field %s (docid=%s)" % (self.field, doc_id))
                debug_print("doc=%s contents=%s" % (doc, contents), 8)
                debug_print("get_doc_terms_and_freqs(%s) => (%s, %s)" % (doc_id, terms, tfs), 7)
                return (terms, tfs)
            terms_enum = tv.iterator(None)
            terms = []
            tfs = []
            for utf8_term in BytesRefIterator.cast_(terms_enum):
                dp_enum = terms_enum.docsAndPositions(None, None)
                if not dp_enum:
                    debug_print("Warning: Problem enumerating document positions for term %s" % utf8_term, 3)
                    continue
                dp_enum.nextDoc()  # note: primes the enum which works only for the current doc
                freq = dp_enum.freq()
                debug_print("type(utf8_term)=%s" % type(utf8_term), 8)
                # note: term returned is Unicode, so codec.encode required before I/O [see print_stderr in tpo_common.py])
                term = utf8_term.utf8ToString()
                debug_print("type(term)=%s" % type(term), 8)
                debug_print("term: %s" % term, 8)
                if SKIP_PLACEHOLDERS and (term.startswith("_ ") or term.endswith(" _")):
                    debug_print("skipping term with placeholders: %s" % term, 7)
                    continue
                terms.append(term.encode('utf-8'))
                tfs.append(freq)
        except:
            if (not IGNORE_EXCEPTIONS):
                tpo.debug_raise()
            debug_print("Warning: Exception during get_doc_terms_and_freqs: %s" % str(sys.exc_info()), 4)
        debug_print("get_doc_terms_and_freqs(%s) => (%s, %s)" % (doc_id, terms, tfs), 7)
        return (terms, tfs)

    def get_all_terms(self):
        """A generator yielding of all terms in the index"""
        # via http://mail-archives.apache.org/mod_mbox/lucene-pylucene-dev/201402.mbox/%3CCAPJ5eE-WM78oZronESNAcurM3azp=ho2XHv-HS4aVr_S5V18Yg@mail.gmail.com%3E
        tpo.debug_format("in get_all_terms()", 6)
        fields = MultiFields.getFields(self.reader)
        terms = fields.terms(self.field)
        terms_enum = terms.iterator(None)
        terms_ref = BytesRefIterator.cast_(terms_enum)
        try:
            while (terms_ref.next()):
                term_value = TermsEnum.cast_(terms_ref)
                term_token = tpo.normalize_unicode(term_value.term().utf8ToString())
                tpo.debug_print("term: %s" % term_token, 7)
                yield term_token
        except StopIteration:
            pass
        except:
            print_stderr("Exception in get_all_terms: %s" % str(sys.exc_info()))
        tpo.debug_format("out get_all_terms()", 6)
        return

    def count_terms(self):
        """Returns total count of terms in index"""
        tpo.debug_format("count_terms()", 6)
        fields = MultiFields.getFields(self.reader)
        terms = fields.terms(self.field)
        terms_enum = terms.iterator(None)
        terms_ref = BytesRefIterator.cast_(terms_enum)
        num_terms = 1
        try:
            while (terms_ref.next()):
                num_terms += 1
        except StopIteration:
            pass
        except:
            print_stderr("Exception in count_terms: %s" % str(sys.exc_info()))
        tpo.debug_format("count_terms() => {num_terms}", 6)
        return (num_terms)

    def get_doc_terms(self, doc_id):
        """Return list of all indexed terms for DOC_ID"""
        # TODO: rework so that term vectors not required
        # TODO: customize for returning number of terms in the document
        gh.assertion(not USE_SOLR)
        (terms, _tfs) = self.get_doc_terms_and_freqs(doc_id)
        debug_print("get_doc_terms(%s) => %s" % (doc_id, terms), 6)
        return (terms)

    def get_docid(self, doc_id):
        """Get document label for DOC_ID to use for output listings (e.g., based on NAME field)"""
        # TODO: rename to get_doc_label once anlyze_tfidf.py updated
        gh.assertion(not USE_SOLR)
        doc_label = doc_id
        if USE_NAME_FOR_ID:
            if not isinstance(doc_id, int):
                doc_id = int(doc_id)
            doc_label = self.searcher.doc(doc_id).get(NAME)
        debug_print("get_docid(%s) => %s" % (doc_id, doc_label), 6)
        return (doc_label)

    def get_tfidf(self, doc_id):
        """Return list of top terms for DOC_ID ranked by TFIDF"""
        # Get list of terms and corresponding frequency for current document
        # note: term vector access based on ./samples/TermPositionVector.py in PyLucene source
        gh.assertion(not USE_SOLR)
        tfidf_info = []
        num_docs = self.get_num_docs()
        try:
            (terms, tfs) = self.get_doc_terms_and_freqs(doc_id)
            if SKIP_UNIGRAMS:
                ## TODO: (terms, tfs) = unzip([(t, f) for (t, f) in zip(terms, tfs) if not " " in t])
                terms_tfs = [(t, f) for (t, f) in zip(terms, tfs) if " " in t]
                terms = [t for (t, _f) in terms_tfs]
                tfs = [f for (_t, f) in terms_tfs]
            num_terms = len(terms)

            # Get (global) document frequency for term and compute TF/IDF
            # TODO: -log(1/dfs[i]) => log(N/dfs[i])
            indices = [i for i in range(len(terms))]
            dfs = [self.doc_freq(terms[i]) for i in indices]
            tfidfs = [tfs[i] * (math.log(num_docs / float(dfs[i])) if (dfs[i] > 0) else 0.0) for i in indices]
            tpo.trace_array(sorted(zip(terms, tfs, dfs, tfidfs), reverse=True, key=lambda t: t[-1]),
                            6, "term info")
            tpo.debug_format("#terms={num_terms}", 4)

            # Sort and get top-N items
            # TODO: filter phrasals subsumed by larger phrases
            sorted_indices = sorted(indices, reverse=True, key=lambda i: tfidfs[i])
            tfidf_info = ["%s:%s" % (terms[i], tpo.round_num(tfidfs[i])) for i in sorted_indices[:MAX_TERMS]]
        except Exception:
            if (not IGNORE_EXCEPTIONS):
                tpo.debug_raise()
            tpo.print_stderr("Warning: Exception during get_tfidf for doc %s: %s" % (doc_id, str(sys.exc_info())))
        if (len(tfidf_info) == 0):
            debug_print("Warning: No terms extracted for document %s" % doc_id)
        return (tfidf_info)

    def perform_query(self, query_text, field=None):
        """Helper function to performs all-words query using specified QUERY_TEXT (in FIELD), returing tuples with (estimated) number of hits and top documents"""
        # TODO: exclude or escape words that are lucene keywords (e.g., "NOT")
        # TODO: escape special characters: \ + - ! ( ) : ^ ] { } ~ * ?
        debug_print("perform_query(%s, [field=%s])" % (query_text, field), 6)
        query_field = field if field else self.field
        num_hits = 0
        try:
            query_text = query_text.strip()
            if PHRASAL and (" " in query_text) and (query_text[0] != '"'):
                query_text = '"' + query_text + '"'
            if USE_SOLR:
                # format: search(q, [kwargs=None])
                ## TODO: scoreDocs = self.solr.search(query_text, **{'df': query_field})
                ## BAD: solr_results = self.solr.search(query_text, **{'df': query_field, 'fl': 'id'})
                params = {
                    # note: df is for default fault, and fl is for (display) field list (thanks, SolR!)
                    ## TODO: specify query override
                    # 'df': query_field, 
                    ## OLD: 'fl': 'id,score',
                    'fl': NAME + ',score',
                    }
                if USE_DISMAX:
                    params['defType'] = "edismax"
                if CONJUNCTIVE:
                    # Use conjuntive query (disjunction is default)
                    params['q.op'] = 'AND'
                tpo.debug_format("params={p}", 5, p=params)
                solr_results = self.solr.search(query_text, **params)
                
                tpo.trace_object(solr_results, 5, "solr_results")
                # note: Solr only returns first 10 hits by default
                scoreDocs = [result['id'] for result in solr_results]
                # TODO: see if way to get total Solr hits
                num_hits = len(scoreDocs)
            else:
                # Treat query as conjunction unless phrasal
                if CONJUNCTIVE and not PHRASAL:
                    query_tokens = re.split(r"\W+", query_text)
                    query_text = " ".join([("+" + t) for t in query_tokens])
                query = QueryParser(Version.LUCENE_CURRENT, query_field,
                                    self.analyzer).parse(query_text)
                tpo.debug_format("Lucene query={query}; query_text='{query_text}'; field={query_field}", 6)
                top_docs = self.searcher.search(query, MAX_HITS)
                num_hits = top_docs.totalHits
                scoreDocs = top_docs.scoreDocs
                
            debug_print("%s returned documents for query '%s'" % (len(scoreDocs), query_text), 5)
            debug_print("%s total matching documents (estimated)" % (num_hits), 5)
            debug_print("scoreDocs=%s" % scoreDocs, 7)
            if RETAIN_DOC_ID:
                doc_names = [str(doc.doc) for doc in scoreDocs]
            else:
                if USE_SOLR:
                    doc_names = scoreDocs
                else:
                    # TODO: convert documents names to UTF-8 (e.g., to get rid of u'...' in output)
                    doc_names = [self.searcher.doc(doc.doc).get(NAME) or "" for doc in scoreDocs]
            if SHOW_TFIDF:
                gh.assertion(not USE_SOLR)
                for doc in scoreDocs:
                    # Display top cases
                    # TODO: use JSON-style format (to facilitate extraction); likewise below
                    debug_print("docid=%s tfidf=%s" % (self.get_docid(doc.doc), self.get_tfidf(doc.doc)), 6)
        except Exception:
            if (not IGNORE_EXCEPTIONS):
                tpo.debug_raise()
            tpo.print_stderr("Warning: Exception during perform_query of %s: %s" % (query_text, str(sys.exc_info())))
            doc_names = []
        result = (num_hits, doc_names)
        debug_print("perform_query(%s) => %s" % (query_text, result), 6)
        return result

    def issue_query(self, query_text, field=None):
        """Performs all-words query using specified TEXT (in FIELD)"""
        (_num_hits, docs) = self.perform_query(query_text, field)
        debug_print("issue_query(%s) => %s" % (query_text, docs), 5)
        return docs

    def get_hit_count(self, query_text, field=None):
        """Returns (estimated) number of documents matching QUERY_TEXT"""
        (num_hits, _docs) = self.perform_query(query_text, field)
        debug_print("get_hit_count(%s) => %s" % (query_text, num_hits), 5)
        return num_hits

    def doc_freq(self, term, cache=True):
        """Returns number of documents that TERM occurs in"""
        # TODO: rename as term_doc_freq???
        debug_print("doc_freq(%s)" % term, 8)
        gh.assertion(not USE_SOLR)
        if cache and term in self.frequencies:
            return (self.frequencies[term])
        try:
            term_object = Term(FIELD, term.strip())
            ## BAD (misleading documentation at http://lucene.apache.org/core/4_0_0/core/org/apache/lucene/index/IndexReader.html)
            ##    num_docs = self.reader.totalTermFreq(term_object)
            num_docs = self.reader.docFreq(term_object)
        except Exception:
            if (not IGNORE_EXCEPTIONS):
                tpo.debug_raise()
            tpo.print_stderr("Warning: Exception during doc_freq of %s: %s" % (term, str(sys.exc_info())))
            num_docs = 0
        if cache:
            self.frequencies[term] = num_docs
        debug_print("doc_freq(%s) => %d" % (term, num_docs), 7)
        return num_docs

#------------------------------------------------------------------------
# Support for distributed processing
# note: this has now been been moostly integrated into remote_dispatching.py

def invoke_distributed(command_line, index_dir, total_num_docs):
    """Divide the task among client workers and then gather results"""
    # TODO: put generic processing in remote_dispatching.py
    # Note: To avoid having to parse the command line, all arguments are given via environment
    # and passed along.
    # TODO: import remote_dispatching above so that environment options made available for --help option
    import remote_dispatching as rd
    import time
    remote_workers = rd.REMOTE_WORKERS.split()
    num_workers = len(remote_workers)
    assert(num_workers > 0)
    dispatcher = rd.RemoteDispatcher()
    # Send index to clients and extract under /tmp
    if not SKIP_INDEX:
        dispatcher.copy_dir_to_workers(index_dir)
    # Reconstruct command line, including environment options with index directory override,
    # and send to clients.
    # TODO: start=i*INCR num=INCR
    environment_spec = dispatcher.get_bash_environment_options(ignore="DISTRIBUTED")
    index_basename = gh.basename(index_dir)
    num_docs = total_num_docs / len(remote_workers)
    command_base = "run_search_table_file_index"
    command_script = tpo.format("/tmp/{command_base}.sh")

    # Invoke commands
    # TODO: use run-specific filenames (e.g., /tmp/remote_search_table_file_index-<PID>.sh)
    this_module = gh.basename(__file__, ".py")

    remote_flag_file = tpo.format("/tmp/{command_base}.done")
    invocation_base = "invoke_remote_search_table_file_index"
    invocation_output_template = "/tmp/{invocation_base}.h{i}.output"
    for i in range(num_workers):
        invocation_script = tpo.format("/tmp/{invocation_base}.sh")
        invocation_output_file = tpo.format(invocation_output_template)
        start_doc = i * num_docs
        end_doc = start_doc + num_docs - 1
        gh.write_lines(command_script, [
            tpo.format("rm -f {remote_flag_file}"),
            # note: /tmp/tohara-1.1 created as part of setup
            tpo.format("{environment_spec} DEBUG_LEVEL=4 PYTHONPATH=/tmp/tohara-1.1 INDEX_DIR=/tmp/{index_basename} START_DOC={start_doc} END_DOC={end_doc} python -m {this_module} - >| {invocation_output_file}  2> {invocation_output_file}.log"),
            tpo.format("touch {remote_flag_file}"),
            ])
        gh.write_lines(invocation_script, [
            tpo.format("source {command_script} &")
            ])
        dispatcher.copy_file_to_worker(i, command_script)
        dispatcher.copy_file_to_worker(i, invocation_script)
        dispatcher.run_command_on_worker(i, tpo.format("source {invocation_script}"))

    # Wait for each host to finish, and then output corresponding result file
    # TODO: record process ID and make sure still running remotely
    num_left = num_workers
    still_running = [True] * num_workers
    time_slept = 0
    while (num_left > 0):
        if (time_slept >= MAX_SLEEP):
            tpo.print_stderr(tpo.format("Error: time out reached ({MAX_SLEEP} seconds) in invoke_distributed"))
            break
        time.sleep(SLEEP_SECONDS)
        time_slept += SLEEP_SECONDS

        # Check each active host for completion, downloading results when reached.
        for i in range(num_workers):
            if still_running[i]:
                flag_found = dispatcher.run_command_on_worker(i, tpo.format("ls {remote_flag_file} 2> /dev/null"))
                if flag_found:
                    still_running[i] = False
                    num_left -= 1
                    remote_output_file = tpo.format(invocation_output_template)
                    dispatcher.copy_file_from_worker(i, remote_output_file)
                    local_output_file = gh.basename(remote_output_file)
                    tpo.debug_format("Output from host {i}: {local_output_file}", 4)
                    print(gh.read_file(local_output_file))
    return

#-------------------------------------------------------------------------------

def  main():
    """Entry point for script"""
    # Check command-line arguments
    # TODO: rework via argparse (e.g., removing nonstandard '-' for index-dir)
    debug_print("sys.argv: %s" % sys.argv, 5)
    if ((len(sys.argv) <= 1) or (sys.argv[1] == "--help")):
        print(__doc__.format(script=sys.argv[0], descriptions=tpo.formatted_environment_option_descriptions()))
        sys.exit()
    index_dir = INDEX_DIR
    if (sys.argv[1] != "-"):
        index_dir = sys.argv[1]

    # Initialize the index lookup
    il = IndexLookup(index_dir)
    tpo.trace_object(il, 7)
    
    # For distributed processing, act as dispatcher and then exit
    if DISTRIBUTED:
        invoke_distributed(sys.argv, index_dir, il.get_num_docs())
        sys.exit()

    # If additional arguments, issue query from remainder of command line
    # Note: if all numeric, treated as document ID's
    # TODO: unify the tfidf listing formatting
    if (len(sys.argv) > 2):
        remainder = sys.argv[2:]
        if SHOW_FREQ:
            gh.assertion(not USE_SOLR)
            if remainder == ["-"]:
                remainder = il.get_all_terms()
            print("term\tfreq")
            for term in remainder:
                print("%s\t%d" % (term, il.doc_freq(term, cache=False)))
        elif (SHOW_TFIDF and all(tpo.is_numeric(arg) for arg in remainder)):
            gh.assertion(not USE_SOLR)
            for doc_id in remainder:
                print("docid=%s tfidf=%s" % (il.get_docid(doc_id), il.get_tfidf(doc_id)))
        else:
            query_text = " ".join(remainder)
            if JUST_HITS:
                result = il.get_hit_count(query_text)
            else:
                doc_names = il.issue_query(query_text)
                result = "\n".join(doc_names)
            if SHOW_TFIDF:
                gh.assertion(not USE_SOLR)
                for doc_id in doc_names:
                    print("docid=%s tfidf=%s" % (il.get_docid(doc_id), il.get_tfidf(doc_id)))
            else:
                print(result)

    # Handle processing when no arguments given
    else:
        # Special case to produce TF/IDF listings for all documents
        if COUNT_TERMS:
            print("Number of terms: %s" % il.count_terms())
        elif SHOW_TFIDF:
            gh.assertion(not USE_SOLR)
            # TODO: rework via PyLucene enumerate-documents API
            num_missing = 0
            all_docs = range(0, il.get_num_docs())
            if REVERSE_ORDER:
                all_docs = sorted(all_docs, reverse=True)
            for doc in all_docs:
                # Get the next TF/IDF listing and do some sanity checks
                tfidfs = il.get_tfidf(doc)
                if tfidfs is None:
                    break
                if (len(tfidfs) == 0):
                    num_missing += 1
                    if (num_missing == MAX_EMPTY):
                        debug_print("Warning: stopping due to %d empty tf/idf listings" % MAX_EMPTY)
                        break

                # Print the info and advance document
                print("docid=%s tfidf=%s" % (il.get_docid(doc), tfidfs))
                doc += 1

        # Otherwise, run interactively
        else:
            il.run()
    return
            
#------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#! /usr/bin/env python
#
# Test(s) for analyze_tfidf.py. 
#
# The main test automates the following steps:
#
#   $ {
#     base=/tmp/test_analyze_tfidf-$$
#     index_dir="$base-index"
#   
#     cut -f1,12 tests/sample-query-log.retrieve_lr_inputs.data >| $base.data
#   
#     SHOW_TFIDF=1 INCLUSION_FIELDS=2  python  -m index_table_file  $base.data  $index_dir >| $index_dir.index.log
#   
#     SHOW_TFIDF=1 python -m search_table_file_index  $index_dir >| $index_dir.search.log
#   
#     INCLUDE_CONTEXT=1 INDEX_DIR=$index_dir SHOW_TFIDF=1 python -m analyze_tfidf  $index_dir.search.log >| $index_dir.analyze.log
#   }
#   
#   $ grep -c biostatistics $base.data
#   1
#   $ calc "log(92/1)"
#   4.52179
#   
#   $ grep tfidf.*biostatistics $index_dir.analyze.log
#   docid=1 tfidf=['biostatistics:4.522']
#   
#   $ grep -c ONCOLOGY $base.data
#   16
#   
#   $ calc "log(92/16)"
#   1.7492
#   
#   $ grep "tfidf=.*oncology" $index_dir.analyze.log | head -1
#   docid=52 tfidf=['oncology:1.749', 'radiation:1.749']
#------------------------------------------------------------------------
# Sample data (columns 1 and 12 of consolidated data file sample):
#
#   impression	k
#   1aa61096dd9a4126b4ecf6efcfd05605	'biostatistics'
#   d982b02353e441568fd3cd24d4719320	'personal-traveling-assistant'
#   H2P2ChGmSgCv1EjSaXoSUQ	'personal-traveling-assistant'
#   ...
#   0Uu62hBlRomLRzVHJumFuQ	'patent attorney'
#   J6dl5f8xQX-fMC0PFBM4sw	'patent attorney'
#------------------------------------------------------------------------
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH="." python tests/test_analyze_tfidf.py
#
# Warning:
# - This is basically the version from the repo misc-scripts.
# - This is mainly included for sake of completeness, because
#   analyze_tfidf.py got added to the repo inadvertantly.
#

import re
import tempfile
import unittest

import glue_helpers as gh
import tpo_common as tpo


class TestIt(unittest.TestCase):
    """Class for testcase definition"""
    script_module_name = "analyze_tfidf"  # name for invocation via 'python -m ...'

    def setUp(self):
        """Per-test initializations: disables tracing of scripts invoked via run(); initializes temp file name (With override from environment)."""
        # Note: By default, each test gets its own temp file.
        tpo.debug_print("setUp()", 4)
        gh.disable_subcommand_tracing()
        self.base = tpo.getenv_text("TEMP_FILE", tempfile.NamedTemporaryFile().name)
        self.index_dir = self.base + "-index"
        return

    def run_script(self, env_options, arguments):
        """Runs the script setting ENV_OPTIONS and passing ARGUMENTS"""
        tpo.debug_format("run_script({env_options}, {arguments})", 5)
        # Usage: ENVVAR1=VAL1 ... analyze_tfidf.py [-h] [--verbose] [filename]
        output = gh.run("{env_options} python  -m {self.script_module_name}   {arguments}")
        # Make sure no python or bash errors
        # examples: "SyntaxError: invalid syntax", "bash: python: command not found"
        self.assertTrue(not re.search("(\S+Error:)|(no module)|(command not found)", output.lower()))
        return (output)

    def test_data_file(self):
        """Tests results over a known data file"""
        tpo.debug_print("test_data_file()", 4)
        data_file = gh.resolve_path("sample-query-log.retrieve_lr_inputs.data")

        # Extract impresssion IDs and query keywords
        base = self.base
        gh.run("cut -f1,12 {data_file} >| {self.base}.data")
        self.assertTrue(len(gh.read_lines(base + ".data")) == 92)

        # Index the keywords from sample query log
        output = gh.run("SHOW_TFIDF=1 INCLUSION_FIELDS=2  INDEX_DIR={self.index_dir} python  -m index_table_file  {base}.data")
        self.assertTrue(re.search("commiting.*done", output))

        # Create TF/IDF listing for documents
        gh.run("SHOW_TFIDF=1 INDEX_DIR={self.index_dir}  python -m search_table_file_index - >| {base}.search.log")
        output = gh.read_file(base + ".search.log")
        # ex: doc=0 tfidf=['k:3.829']
        self.assertTrue(re.search("docid=0.*tfidf.*'k:", output))
        # ex: doc=91 tfidf=['attorney:2.124', 'patent:2.037']
        self.assertTrue(re.search("docid=91.*tfidf=.*attorney.*patent", output))

        # Perform posthoc TF/IDF analysis
        gh.write_lines(base + ".exclude", [ "attorney" ])
        ## OLD: output = gh.run("INCLUDE_CONTEXT=1 EXCLUSION_FILE={base}.exclude INDEX_DIR={self.index_dir} SHOW_TFIDF=1  python -m analyze_tfidf  {base}.search.log")
        ## BAD: output = self.run_script("INCLUDE_CONTEXT=1 EXCLUSION_FILE={base}.exclude INDEX_DIR={self.index_dir} SHOW_TFIDF=1", "{base}.search.log")
        output = self.run_script(tpo.format("INCLUDE_CONTEXT=1 EXCLUSION_FILE={base}.exclude INDEX_DIR={self.index_dir} SHOW_TFIDF=1"), tpo.format("{base}.search.log"))
        # ex: doc=[assistant, personal, traveling]
        self.assertTrue(re.search("assistant.*personal.*traveling", output))
        # ex: 92 documents
        # ex: 393 keywords; 27 distinct
        # ex: mean mean TF/IDF: 4.268
        self.assertTrue(re.search("92 documents", output))
        self.assertTrue(re.search("[0-9][0-9][0-9] keywords.* [0-9][0-9] distinct", output))
        self.assertTrue(re.search("mean TF/IDF.* [2-6]\.", output))
        return
        
    def tearDown(self):
        """Per-test cleanup: deletes temporary file unless during detailed debugging"""
        tpo.debug_print("tearDown()", 4)
        if (not tpo.detailed_debugging()):
            gh.run("rm -rvf {self.base}*")
        return

#------------------------------------------------------------------------

if __name__ == '__main__':
    unittest.main()




#! /usr/bin/env python3
#
# Miscellaneous functions not suitable for other modules (e.g., system.py).
#
# TODO:
# - Separate list related functions (e.g., as list_utils.py).
#

"""Misc. utility functions"""

# Standard packages
import datetime
from difflib import ndiff
import inspect
import math
import os
import random
import re
import sys
import time
import json
import yaml
import csv
from types import ModuleType
from typing import Any, Optional
from contextlib import ContextDecorator

# Installed packages
## NOTE: made dynamic due to import issue during shell-script repo tests
## TODO3: import more_itertools

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import dummy_app
from mezcla.my_regex import my_re
from mezcla import system
from mezcla import text_utils

# Constants
ELLIPSIS = "\u2026"                 # Horizontal Ellipsis
TYPICAL_EPSILON = system.getenv_float(
    "TYPICAL_EPSILON", 1e-6,
    description="Traditional floating-point negligible difference")
VALUE_EPSILON = system.getenv_float(
    "VALUE_EPSILON", 1e-3,
    description="Epsilon for informal floating-point comparison")
debug.assertion(TYPICAL_EPSILON < VALUE_EPSILON)
RANDOM_SEED = system.getenv_integer(
    "RANDOM_SEED", 15485863,
    ## TEST: "RANDOM_SEED", 15485863 ** 2 - 1,
    description="Integral seed for random number generation: 0 for default")


def transitive_closure(edge_list):
    """Computes transitive close for graph given by EDGE_LIST (i.e., makes indirect links explicit)"""
    # ex: transitive_closure([(1,2),(2,3),(3,4)]) => set([(1, 2), (1, 3), (1, 4), (2, 3), (3, 4), (2, 4)])
    # notes; based on https://stackoverflow.com/questions/8673482/transitive-closure-python-tuples
    closure = set(edge_list)
    while True:
        new_relations = set((x, w) for x, y in closure for q, w in closure if q == y)

        closure_until_now = closure | new_relations
        if closure_until_now == closure:
            break

        closure = closure_until_now

    return closure


def read_tabular_data(filename):
    """Reads table with (unique) key and tab-separated value. 
    Note: key made lowercase"""
    debug.trace_fmtd(4, "read_tabular_file({f})", f=filename)
    table = {}
    with system.open_file(filename) as f:
        for (i, line) in enumerate(f):
            line = system.from_utf8(line)
            items = line.split("\t")
            if len(items) == 2:
                debug.assertion(items[0].lower() not in table)
                table[items[0].lower()] = items[1]
            else:
                debug.trace_fmtd(4, "Ignoring item w/ unexpected format at line {num}",
                                 num=(i + 1))
    ## debug.trace_fmtd(7, "table={t}", t=table)
    debug.trace_values(7, table, "table")
    return table


def extract_string_list(text, strip_empty=False, allow_numeric_ranges=False):
    """Extract a list of values from text using whitespace and/or commas as delimiters.
    With STRIP_EMPTY, empty-string items are ignored.
    """
    # EX: extract_string_list("1  2,3") => ["1", "2", "3"]
    # TODO: Add support for quoted values to allow for embedded spaces
    trimmed_text = my_re.sub(r"\s+", " ", text.strip())
    trimmed_text = my_re.sub(r"([^,]) ", r"\1,", trimmed_text)
    values = my_re.split(", ?", trimmed_text) if trimmed_text else []
    #
    if allow_numeric_ranges:
        new_values = []
        for value in values:
            if my_re.search(r"^(\d+)-(\d+)", value):
                new_values += list(str(v) for v in
                                   range(int(my_re.group(1)), int(my_re.group(2))+ 1))
            else:
                new_values.append(value)
        values = new_values
    #
    if strip_empty:
        values = [v for v in values if v]
    debug.trace_fmtd(5, "extract_string_list({t!r}) => {v!r}", t=text, v=values)
    return values
#
# EX: extract_string_list("1,", strip_empty=True) => ["1"]
# EX: extract_string_list("1-3", allow_numeric_ranges=True) => ["1", "2", "3"]


def is_prime(num):
    """Moderately efficient function for testing whether a number is prime
    Notes:
    - Based on https://en.wikipedia.org/wiki/Primality_test.
    - The intuition is that primes > 3 are of form (6k +/- 1).
    """
    ## EX: FIRST_100_PRIMES = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97, 101, 103, 107, 109, 113, 127, 131, 137, 139, 149, 151, 157, 163, 167, 173, 179, 181, 191, 193, 197, 199, 211, 223, 227, 229, 233, 239, 241, 251, 257, 263, 269, 271, 277, 281, 283, 293, 307, 311, 313, 317, 331, 337, 347, 349, 353, 359, 367, 373, 379, 383, 389, 397, 401, 409, 419, 421, 431, 433, 439, 443, 449, 457, 461, 463, 467, 479, 487, 491, 499, 503, 509, 521, 523, 541]; (len(FIRST_100_PRIMES) == 100)
    ## EX: all([is_prime(n) for n in FIRST_100_PRIMES])
    ## EX: all([(not is_prime(n)) for n in range(FIRST_100_PRIMES[-1])  if n not in FIRST_100_PRIMES])
    ##
    debug.trace_fmt(5, "in is_prime({n})", n=num)
    is_prime_num = True

    # First, check primes below 4 (only 2)
    if (num <= 3):
        is_prime_num = (num > 1)
        if not is_prime_num:
            debug.trace_fmt(4, "{n} not prime as less than 3 and not 2.", n=num)
        return is_prime_num

    # Next, make sure not divisible by 2 or 3
    if ((num % 2 == 0) or (num % 3 == 0)):
        debug.trace_fmt(4, "{n} not prime as divisible by 2 or 3.", n=num)
        return False

    # Otherwise, check (6k +/- 1) values to see if divisible by 2 or 3,
    # stopping when value exceeds sqrt(n).
    last_possible = math.ceil(math.sqrt(num))
    i = 5
    while (i <= last_possible):
        if (((num % i) == 0) or ((num % (i + 2)) == 0)):
            debug.trace_fmt(4, "{n} not prime as (6k +/- 1) divisible by 2 or 3 for some k.", n=num)
            is_prime_num = False
            break
        i += 6

    debug.trace_fmt(5, "is_prime({n}) => {ip}", n=num, ip=is_prime_num)
    return is_prime_num


def prime_factorization(num):
    """Return list of primne factors for NUM"""
    ## EX: prime_factorization(123) => [3, 41]
    ## EX: prime_factorization(127) => [127]
    i = 2
    factors = []
    while i * i <= num:
        if num % i:
            i += 1
        else:
            num //= i
            factors.append(i)
    if num > 1:
        factors.append(num)
    debug.trace(6, f"prime_factorization({num}) => factors")
    return factors


def fibonacci(max_num):
    """Returns Fibonacci sequence with numbers less than MAX_NUM"""
    # EX: fibonacci(10) => [0, 1, 1, 2, 3, 5, 8]
    a, b = 0, 1
    sequence = []
    while (a < max_num):
        sequence.append(a)
        a, b = b, (a + b)
    debug.trace_fmt(9, "fibonacci({m}) => {s}", m=max_num, s=sequence)
    return sequence


def sort_weighted_hash(weights, max_num=None, reverse=None):
    """sorts the entries in WEIGHTS hash, returns list of (key, freq) tuples.
    Note: sorted in REVERSE order by default"""
    if max_num is None:
        max_num = len(weights)
    if reverse is None:
        reverse = True
    sorted_keys = sorted(weights.keys(), reverse=reverse, 
                         key=lambda k: weights[k])
    top_values = [(k, weights[k]) for k in sorted_keys[:max_num]]
    debug.trace_fmt(5, "sort_weighted_hash(_, [max={m}], [rev={r}) => {t}",
                    m=max_num, r=reverse, t=top_values)
    return top_values


def unzip(iterable, num=None):
    """"Inverse of zip operation: returns n lists of i-th elements of input list of tuples. The optional NUM argument ensures that that many values returned (in case of empty input lists)"""
    # See https://stackoverflow.com/questions/19339/transpose-unzip-function-inverse-of-zip.
    # EX: unzip(zip([1, 2, 3], ['a', 'b', 'c'])) => [[1, 2, 3], ['a', 'b', 'c']]
    # EX: unzip(zip([], []), 2) => [[], []]
    result = [list(tupl) for tupl in zip(*iterable)]
    if not result and num:
        result = [[] for _i in range(num)]
    return result


def get_current_frame():
    """Return stack frame (i.e., for caller)"""
    frame = inspect.currentframe().f_back
    debug.trace_fmt(6, "get_current_frame() => {f}", f=frame)
    return frame


def eval_expression(expr_text, frame=None):
    """Evaluate EXPRESSION_given_by_TEXT, using FRAME
    Note: Uses caller's frame if not given. This should only be used
    when the expression involve local variables (see extract_matches.py
    in shell-scripts repo).
    """
    # EX: eval_expression("len([123, 321]) == 2")
    result = None
    try:
        if frame is None:
            frame = get_current_frame()
            # Note: need to get caller's frame
            frame = frame.f_back
        # pylint: disable=eval-used
        result = eval(expr_text, frame.f_globals, frame.f_locals)
    except:
        debug.trace_fmt(5, "Exception during eval_expression({expr}): {exc}",
                        expr=expr_text, exc=sys.exc_info())
    debug.trace_fmt(7, "eval_expression({expr}) => {r}",
                    expr=expr_text, r=result)
    return result


def trace_named_object(level, object_name, caller_frame=None):
    """Trace OBJECT_given_by_NAME
    Note: ***use debug.trace_expr instead ***"""
    # EX: trace_named_object(4, "sys.argv") is None
    if caller_frame is None:
        caller_frame = get_current_frame().f_back
    debug.trace_object(level, eval_expression(object_name,
                                              frame=caller_frame),
                       label=object_name)
    return


def trace_named_objects(level, list_text):
    """Trace objects in LIST_text
    Note: *** use debug.trace_expr instead ***"""
    # EX: trace_named_objects(4, "[len(sys.argv), sys.argv]") is None
    debug.assertion(re.search(r"^\[.*\]$", list_text))
    frame = get_current_frame().f_back
    for name in text_utils.extract_string_list(list_text[1:-1]):
        trace_named_object(level, name, caller_frame=frame)
    return


def exactly1(items):
    """Whether one and only one if ITEMS is true"""
    ## TEMP: does import here to avoid problem with mezcla package install
    import more_itertools               # pylint: disable=import-outside-toplevel
    ok = more_itertools.exactly_n(items, 1)
    debug.trace(4, f"exactly1({items}) => {ok}")
    return ok 


def string_diff(text1, text2):
    """Return diff-style comparison of TEXT1 and TEXT2 with an empty string used for equality"""
    debug.trace(6, f"string_diff({text1}, {text2})")
    # EX: string_diff("one\ntwo\nthree\nfour", "one\ntoo\ntree\nfour") => "  one\n< two\n…  ^\n> too\n…  ^\n< three\n…  -\n> tree\n  four\n"

    # Perform comparison
    diff_result = "n/a"
    try:
        lines1 = (text1 + "\n").splitlines(keepends=True)
        lines2 = (text2 + "\n").splitlines(keepends=True)
        diff_result = "".join(ndiff(lines1, lines2))
    except:
        system.print_exception_info("string_diff compare")

    # Convert to diff-style output, using:
    # < and > instead of - and +
    # … instead of ?
    try:
        diff_result = re.sub(r"^- ", "< ", diff_result, flags=re.MULTILINE)
        diff_result = re.sub(r"^\+ ", "> ", diff_result, flags=re.MULTILINE)
        diff_result = re.sub(r"^\? ", f"{ELLIPSIS} ", diff_result, flags=re.MULTILINE)
    except:
        system.print_exception_info("string_diff postprocess")

    debug.trace(6, f"string_diff() => {diff_result}")
    return diff_result


def elide_string_values(obj, depth=0, max_len=None):
    """Elide the values of all strings in OBJ, which can either be a scalar, a list or a hash"""
    if max_len is None:
        max_len = 32
    MAX_DEPTH = 10
    if isinstance(obj, list) and (depth < MAX_DEPTH):
        obj = [elide_string_values(v, depth=(1 + depth), max_len=max_len) for v in obj]
    elif isinstance(obj, dict) and (depth < MAX_DEPTH):
        obj = {k: elide_string_values(obj[k], depth=(1 + depth), max_len=max_len) for k in obj.keys()}
    elif isinstance(obj, str):
        obj = gh.elide(obj, max_len=max_len)
    else:
        pass
    return obj


def is_close(value1, value2, epsilon=VALUE_EPSILON):
    """Whether VALUE1 and VALUE2 are close (i.e., absolute difference <= epsilon)"""
    # See https://stackoverflow.com/questions/35324893/using-math-isclose-function-with-values-close-to-0
    ## EX: is_close(1.001, 1.002, epsilon=.005)
    ## EX: (not is_close(1.001, 1.002, epsilon=.0005))
    result = math.isclose(value1, value2, abs_tol=epsilon)
    debug.trace(6, f"is_close({value1}, {value2}, [eps={epsilon}]) => {result}")
    return result


def get_formatted_date(date=None, fmt=None, sep=None, timestamp=None):
    """Return (today's) date in DD MMM YY format (e.g., 10 oct 22)
    Note: uses optional SEP between components or FMT
    Use TIMESTAMP to specify date as integer
    """
    ## EX: get_formatted_date(datetime.datetime.fromtimestamp(0)) => "31 dec 69"
    ## TODO3: rename as get_formatted_date
    in_date = date
    if timestamp is not None:
        debug.assertion(date is None)
        date = datetime.datetime.fromtimestamp(timestamp)
    if date is None:
        date = datetime.date.today()
    if sep is None:
        sep = " "
    if fmt is None:
        fmt = f"%d{sep}%b{sep}%y"
    try:
        result = date.strftime(fmt).lower()
    except:
        system.print_exception_info("get_today_ddmmmyy")
        result = "???"
    debug.trace(6, f"get_formatted_date({in_date}, ts={timestamp}) => {result}")
    return result

def get_date_ddmmmyy(date=None):
    """Return (today's) date in DDMMMYY format (e.g., 10oct22)"""
    ## EX: get_date_ddmmmyy(datetime.datetime.fromtimestamp(0)) => "31dec69"
    result = get_formatted_date(date).replace(" ", "")
    debug.trace(6, f"get_date_ddmmmyy({date}) => {result}")
    return result

def parse_timestamp(ts: str, truncate=None, utc=None) -> datetime.datetime:
    """Parse timestamp in ISO 8601 format (e.g., 2023-10-06T04:03:27.1271706Z)
    Note: The timestamp is truncated to micrososeconds unless TRUNCATE is false; and, it is optionally converted to UTC
    """
    # EX: parse_timestamp("2023-10-06T04:03:27.1271706Z") => datetime.datetime(2023, 10, 6, 4, 3, 27, 127170)
    # Note: See https://stackoverflow.com/questions/6207365/working-with-high-precision-timestamps-in-python
    # Code based on ChatGPT suggestion

    # Truncate to microsecond precision
    in_ts = ts
    if truncate is None:
        truncate = (len(ts.split(".")[-1]) > 7)
    if truncate:
        ts = ts[:26] + 'Z'

    # Convert result and change to UTC if desired
    result = datetime.datetime.strptime(ts, '%Y-%m-%dT%H:%M:%S.%fZ')
    if utc:
        result = result.replace(tzinfo=datetime.timezone.utc)
    debug.trace(7, f"parse_timestamp({in_ts}, [{truncate}]) =>  {result}")
    
    return result


def add_timestamp_diff(in_filename, out_filename, prefix=False):
    """Add timestamp difference to each occurrence from IN_FILENAME based on previous occurrence, saving to OUT_FILENAME
    If PREFIX, then the difference is added to start of line, otherwise after timestamp
    """
    # TODO3: isolate as separate utility?
    new_lines = []
    last_time = None
    for line in system.read_lines(in_filename):
        # Check for ISO 8601 timestamp (e.g., 2023-10-06T04:02:36.5228822Z)
        new_line = line
        if my_re.search(r"(\d{4}-\d{1,2}-\d{1,2}T\d{2}:\d{2}:\d{2}\.\d{6,}Z)", line):
            timestamp = my_re.group(1)
            try:
                new_time = parse_timestamp(timestamp)
            except:
                new_time = last_time

            # Compute delta in microsseconds
            time_diff = "0"
            microsec = "\u00B5" + "s"        # U+00B5 (µ)
            if last_time:
                time_diff = "+" + str((new_time - last_time).total_seconds() * 1e6) + microsec
            if prefix:
                new_line = time_diff + "\t" + line
            else:
                new_line = line.replace(timestamp, f"{timestamp} [{time_diff}]")
            last_time = new_time
        new_lines.append(new_line)

    # Output revised lines
    system.write_lines(out_filename, new_lines)


def random_int(min_value=None, max_value=None):
    """Returns random integer in range [MIN_VALUE, MAX_VALUE]
    Note: defaults to [0, sys.maxsize]
    """
    if min_value is None:
        min_value = 0
    if max_value is None:
        max_value = sys.maxsize
    result = random.randint(min_value, max_value)
    debug.trace(6, f"random_int() => {result}")
    return result


def random_float(min_value=None, upper_bound=None):
    """Returns random float in range [min_value, upper_bound)
    Note: defaults to [0, 1).
    """
    if min_value is None:
        min_value = 0
    if upper_bound is None:
        upper_bound = 1
    result = (min_value + (random.random() * (upper_bound - min_value)))
    debug.assertion(result < upper_bound)
    debug.trace(6, f"random_float() => {result}")
    return result


def time_function(func, *args, **kwargs):
    """Time invocation of FUNC, optionally supplied with ARGS and KWARGS"""
    ## EX: is_close(time_function(time.sleep, 5), 5000, epsilon=1)
    start = time.time()
    func(*args, **kwargs)
    end = time.time()
    ms = round(1000.0 * (end - start), 3)
    return ms


def get_class_from_name(class_name, module_name=None):
    """Return class with CLASS_NAME
    Optionally specifies MODULE_NAME containing the class, if not __name__
    """
    # EX: import mezcla.text_processing; "noun_phrases" in dir(get_class_from_name("TextProc", module_name="mezcla.text_processing"))
    # EX: not get_class_from_name("TextProc")
    # See https://stackoverflow.com/questions/1176136/convert-string-to-python-class-object
    if module_name is None:
        module_name = __name__
    class_object = getattr(sys.modules[module_name], class_name, None)
    debug.trace(6, f"get_class_from_name({class_name}, [{module_name}])) => {class_object}")
    return class_object


def convert_file_to_instances(input_file, module_name, class_name, field_names,
                              fmt=None):
    """Converts input file with array of records into a list of class instances.
    Note: Supports mezcla_to_standard mapping table loading.
    Warning: Future support will just be in terms of reading and evaluation python
    text-based data, which is the "py-data" format used below.
    
    Args:
        input_file: Path to input file (json, yaml, csv)
        module_name: Module containing the class definition
        class_name: Name of the class to instantiate
        field_names: List of field names to use for class constructor
        fmt: Optional format override ('json', 'yaml', 'csv', 'py-data'). If None, inferred from extension.
    
    Returns:
        List of class instances
    """
    # pylint: disable=exec-used,eval-used

    # Detect format if not specified
    if not fmt:
        ext = input_file.lower().split('.')[-1]
        fmt = ext

    # Initialize class from module
    exec(f"from {module_name} import *")          
    actual_class = get_class_from_name(class_name, module_name)

    # Read data based on format
    data = []
    if fmt == 'json':
        data = json.loads(system.read_file(input_file))
    elif fmt == 'yaml':
        data = yaml.safe_load(system.read_file(input_file))
    elif fmt == 'csv':
        debug.trace(4, "Warning: convert_file_to_instances problematic with CSV files")
        # Use CSV DictReader to get list of dicts
        with system.open_file(input_file) as f:
            reader = csv.DictReader(f)
            data = list(reader)    
    elif fmt == 'py-data':
        try:
            data = eval(system.read_file(input_file))
        except:
            system.print_exception_info(f"evaluation of {input_file!r}")
    else:
        raise ValueError(f"Unsupported format: {fmt}")

    # Convert records to instances
    instances = []
    for record in data:
        if fmt == 'py-data':
            ## TODO4: rework to put special case outside of loop
            debug.assertion(isinstance(record, actual_class))
            instances.append(record)
            continue
        class_args = []
        for field in field_names:
            # Handle both dict and object-like records
            value = record.get(field, "None")
            # Only eval non-string values from JSON/YAML
            ## TODO3: clarify intention
            ## TEST: if (fmt == 'csv') and isinstance(value, str):
            if fmt in ('json', 'yaml') and isinstance(value, str):
                try:
                    value = eval(value)
                except:
                    system.print_exception_info(f"evaluation of {field} value {value!r}")
            elif (fmt == 'py-data'):
                pass
            else:
                debug.trace(5, f"FYI: Not evaluating {field} value {value!r}")
            class_args.append(value)
        new_inst = actual_class(*class_args)
        instances.append(new_inst)

    return instances    

def convert_json_to_instance(json_file, module_name, class_name, field_names):
    """Converts JSON_FILE with array of dicts into a list of instances for CLASS_NAME, where each of FIELD_NAMES is used in the class invocation. The MODULE_NAME is used to import the class definition.
    """
    return convert_file_to_instances(json_file, module_name, class_name, field_names, fmt='json')

def convert_yaml_to_instance(yaml_file, module_name, class_name, field_names):
    """Converts YAML_FILE with array of dicts into a list of instances for CLASS_NAME, where each of FIELD_NAMES is used in the class invocation. The MODULE_NAME is used to import the class definition.
    """
    return convert_file_to_instances(yaml_file, module_name, class_name, field_names, fmt='yaml')

def convert_csv_to_instance(csv_file, module_name, class_name, field_names):
    """Converts CSV_FILE with array of dicts into a list of instances for CLASS_NAME, where each of FIELD_NAMES is used in the class invocation. The MODULE_NAME is used to import the class definition.
    """
    return convert_file_to_instances(csv_file, module_name, class_name, field_names, fmt='csv')


def convert_python_data_to_instance(python_data_file, module_name, class_name, field_names):
    """Converts PYTHON_DATA_FILE into list of instances
    Note: see convert_json_to_instance for details of other arguments"""
    return convert_file_to_instances(python_data_file, module_name, class_name, field_names, fmt='py-data')


def apply_numeric_suffixes(text: str, just_once=False) -> str:
    """Converts numbers in TEXT to use K, M, G, T, etc. suffixes.
    Note: Optionally applies JUST_ONCE per line."""
    # Note: used in shell-scripts tomohara-aliases.bash
    # EX: apply_numeric_suffixes("1024 1572864 1073741824") => "1K 1.5M 1G"
    # TODO2: preserve spacing in text (e.g., splitlines quirks)
    debug.trace(5, f"apply_numeric_suffixes({text!r}, [once={just_once})")
    suffixes = "_KMGTPE"                # _ is placeholder for no suffix
    new_text = ""
    for l, line in enumerate(text.splitlines()):
        max_count = len(line) / 3
        count = 0
        # note: uses negative look ahead to avoid conversion in decimals (e.g., 1023.5);
        # also, uses non-greedy search to support just_once (i.e., target first case even if smaller)
        # TODO3: exclude leading question marks (e.g., ?00000000?)
        while (my_re.search(r"^(.*?)\b(\d{4,19})\b(?!\.)(.*)", line)):
            count += 1
            if count > max_count:
                debug.trace(6, f"max attempt count reached ({max_count})")
                break
            (pre, numeric, post) = my_re.groups()
            (new_num, suffix) = ("", "")
            num = float(numeric)
            if num > 0:
                try:
                    power = int(math.log(num) / math.log(1024))
                    new_num = system.round_as_str(num / (1024 ** power))
                    # note: accounts for quirk in rounding (e.g., stripping N.000)
                    # TODO2: ignore special cases like 0000 (n.b., due to \d{4,N} regex)
                    new_num = my_re.sub(r"\.0+$", "", new_num)
                    # note: drops 0's added to due rounding (e.g., 1.500 => 1.5)
                    new_num = my_re.sub(r"(\.[1-9]+)0+$", r"\1", new_num)
                    suffix = suffixes[power]
                    debug.trace_expr(6, numeric, num, power, line)
                except:
                    # note: restore number and add surrounding ?'s to block regex
                    new_num = f"_{numeric}?_"
                    debug.trace_exception(5, f"apply_numeric_suffixes line {l} ({line!r})")
            else:
                ## TODO3: rework pattern matching to exclude 0000, etc.
                new_num = "0K"
            line = pre + str(new_num) + suffix + post
            if just_once:
                debug.trace(6, f"applied just once ({just_once=})")
                break
        new_text += line + os.linesep
    if new_text.endswith(os.linesep) and not text.endswith(os.linesep):
        new_text = my_re.sub(fr"{os.linesep}$", "", new_text)
    debug.trace(5, f"apply_numeric_suffixes() => {new_text!r}")
    return new_text


def apply_numeric_suffixes_stdin(just_once=False):
    """Invokes apply_numeric_suffixes over stdin
    Note: supports Bash alias (see shell-scripts/tomohara-aliases.bash)
    """
    text = dummy_app.read_entire_input()
    # Note: doesn't add newline as normally stdin ends with one, and users
    # might want stdin preserved if not (e.g., for use with `echo -n`).
    print(apply_numeric_suffixes(text, just_once=just_once), end="")

#-------------------------------------------------------------------------------
# Utility class for setting contexts
# NOTE: Based on Grok3
# TODO3: put in new module like python_utils.py

class GlobalSetter(ContextDecorator):
    """A context manager and decorator to temporarily set a module-level global variable.

    This class allows you to temporarily modify a module-level global variable within
    a `with` block or as a decorator, restoring the original value (or removing the
    attribute if it didn't exist) upon exit.

    Args:
        module (ModuleType): The module containing the global variable (e.g., `debug`).
        name (str): The name of the global variable (e.g., 'trace_level').
        value (Any): The temporary value to set for the global variable.

    Example:
        >>> import debug
        >>> def trace_values():
        ...     print(debug.trace_level)
        >>> debug.trace_level = 3
        >>> trace_values()  # Prints: 3
        >>> with GlobalSetter(debug, 'trace_level', 6):
        ...     trace_values()  # Prints: 6
        >>> trace_values()  # Prints: 3

    Note:
        - This context manager is not thread-safe. Use synchronization mechanisms
          (e.g., locks) if modifying globals in a multithreaded environment.
        - If the module is reloaded during the context, the global variable may be reset.
    """

    def __init__(self, module: ModuleType, name: str, value: Any) -> None:
        """Initialize the GlobalSetter with the module, attribute name, and temporary value.

        Args:
            module (ModuleType): The module containing the global variable.
            name (str): The name of the global variable to modify.
            value (Any): The temporary value to set for the global variable.
        """
        self.module: ModuleType = module
        self.name: str = name
        self.value: Any = value
        self.original_value: Optional[Any] = None

    def __enter__(self) -> 'GlobalSetter':
        """Enter the context, saving the original value and setting the new value.

        Returns:
            GlobalSetter: The context manager instance (for use in `with` blocks).

        Saves the original value of the module attribute (or None if it doesn't exist)
        and sets the attribute to the specified temporary value.
        """
        self.original_value = getattr(self.module, self.name, None)
        setattr(self.module, self.name, self.value)
        return self

    def __exit__(self, exc_type: Optional[type], exc_val: Optional[Exception], 
                 exc_tb: Optional[Any]) -> None:
        """Exit the context, restoring the original value or removing the attribute.

        Args:
            exc_type (Optional[type]): The type of the exception raised, if any.
            exc_val (Optional[Exception]): The exception instance raised, if any.
            exc_tb (Optional[Any]): The traceback of the exception, if any.

        Restores the original value of the module attribute if it existed, or removes
        the attribute if it was newly created during the context.
        """
        if self.original_value is None:
            delattr(self.module, self.name)
        else:
            setattr(self.module, self.name, self.original_value)

#-------------------------------------------------------------------------------

def init():
    """MOdule initialization"""
    if RANDOM_SEED:
        random.seed(RANDOM_SEED)


init()

#-------------------------------------------------------------------------------

def main(args):
    """Supporting code for command-line processing"""
    debug.trace_fmtd(6, "main({a})", a=args)
    system.print_stderr("Warning: not intended for direct invocation")
    return

if __name__ == '__main__':
    main(sys.argv)



#! /usr/bin/env python
#
# Test(s) for search_table_file_index.py
#
# Notes:
# - Fill out the TODO's through the file.
# - This can be run as follows:
#   $ PYTHONPATH="." python tests/test_search_table_file_index.py
#
# Warning:
# - This is basically the version from the repo misc-scripts.
# - This is mainly included for sake of completeness, because
#   analyze_tfidf.py got added to the repo inadvertantly.
# - In addition, it is used in test_format_profile.py because,
#   the script will likely not change much at all.
#

# Standard modules
import re
import sys
import tempfile
import unittest

# Installed modules
import pytest

# Local modules
from mezcla import glue_helpers as gh
from mezcla import tpo_common as tpo
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:               module object (e.g., <module 'mezcla.main' ...>)
#    TestIt.script_module:     dotted module path (e.g., "mezcla.main")
THE_MODULE = None
try:
    import mezcla.archive.search_table_file_index as THE_MODULE
except:
    system.print_exception_info("search_table_file_index import") 


class TestIt(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    def setUp(self):
        """Per-test initializations: disables tracing of scripts invoked via run(); initializes temp file name (With override from environment)."""
        # Note: By default, each test gets its own temp file.
        tpo.debug_print("setUp()", 4)
        gh.disable_subcommand_tracing()
        self.temp_file = tpo.getenv_text("TEMP_FILE", tempfile.NamedTemporaryFile().name)
        # Get rid of any files left from previous run (e.g., with detailed debugging)
        gh.run("rm -rvf {self.temp_file}*")
        return

    ## OLD:
    ## def run_script(self, env_options, index_dir, arguments):
    ##     """Runs the script over the INDEX_DIR, passing ENV_OPTIONS"""
    ##     tpo.debug_format("run_script({env_options}, {index_dir}, {arguments})", 5)
    ##     # Usage: ENVVAR1=VAL1 ... search_table_file_index.py [index_dir [query_word1 ...]
    ##     output = gh.run("{env_options} python  -m {self.script_module_name}   {index_dir}  {arguments}")
    ##     # Make sure no python or bash errors
    ##     # examples: "SyntaxError: invalid syntax", "bash: python: command not found"
    ##     self.assertTrue(not re.search("(\S+Error:)|(no module)|(command not found)", output.lower()))
    ##     return (output)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_data_file(self):
        """Tests results over a known data file"""
        tpo.debug_print("test_data_file()", 4)
        index_dir = self.temp_file + "-index"
        # Index the sample data file
        # note: contents field stored for TF/IDF check but no term vectors included
        data_file = gh.resolve_path("sample-query-log.retrieve_lr_inputs.data")
        output = gh.run("SHOW_TFIDF=0 STORE_CONTENTS=1 python -m index_table_file {data_file} {index_dir}")
        self.assertTrue(re.search("commiting.*done", output))

        # Test different types of queries
        # TODO: fix the query counts (or use ranges to make more robust)
        #
        # All-words search for "windows nt" returns 89 documents
        output = self.run_script("MAX_HITS=100", index_dir, "windows nt")
        self.assertTrue(len(output.split("\n")) == 89)
        # Phrasal search yields one less document
        output = self.run_script("PHRASAL=1 MAX_HITS=100", index_dir, "windows nt")
        self.assertTrue(len(output.split("\n")) == 88)
        # No match in 'name' field
        output = self.run_script("FIELD=name", index_dir, "windows nt")
        self.assertTrue(len(output) == 0)
        # The number 62 should occur 11 times
        output = self.run_script("", index_dir, "62")
        self.assertTrue(len(output.split("\n")) == 11)
        # There should an error in TF/IDF listing (requires stored field).
        # ex: "Error: No term vectors for field contents (docid=62)"
        output = self.run_script("SHOW_TFIDF=1", index_dir, "62")
        self.assertTrue(re.search("No term vectors", output))

        # Re-index the sample data file
        output = gh.run("SHOW_TFIDF=1 STORE_CONTENTS=1 python -m index_table_file {data_file} {index_dir}")
        self.assertTrue(re.search("commiting.*done", output))
        # There should be one TF/IDF result for document 62, with 'crl' having weight 4.9
        # ex "docid=62 tfidf=['clr:4.904', 'net:4.904', 'wzbyukewqfwo7x9neg9ydq:4.522', '0.022744695:3.829', '30.873058:3.829', '37569:3.829', '491868535:3.829', '28:2.912', '5:2.773', '9hq:2.730']"
        output = self.run_script("SHOW_TFIDF=1", index_dir, "62")
        self.assertTrue(len(output.split("\n")) == 1)
        self.assertTrue(re.search("\['clr:4\.9", output))
        return
 
    def tearDown(self):
        """Per-test cleanup: deletes temporary file unless during detailed debugging"""
        tpo.debug_print("tearDown()", 4)
        if (not tpo.detailed_debugging()):
            gh.run("rm -rvf {self.temp_file}*")
        return


#------------------------------------------------------------------------

tpo.debug_format("__name__={__name__} sys.argv={sys.argv}", 6)
if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python3
#
# Miscellaneous tests not tied to particular module.
#
# Note:
# - This is uses to check for enforce some development
#   -- A test exists for each module (e.g., tests/test_fubar.py for ./fubar.py).
#   -- Python files have execute permissions (e.g., chmod ugo+x).
#
# TODO2: Run once a week or so (e.g., to help catch poor test stubs like
# test_train_language_model.py)!
#
# TODO3:
# - Check for common pylint issues (n.b., not nitpicking ones like spacing).
#

"""Miscellaneous/non-module tests"""

# Standard packages
import os

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla.tests.common_module import SKIP_SLOW_TESTS, SKIP_SLOW_REASON

# Constants
LINE_IMPORT_PYDANTIC = "from pydantic import validate_call\n"
## TODO1: don't hardcode /tmp (see below)
## OLD: OUTPUT_PATH_PYDANTIC = "/tmp/temp_"
DECORATOR_VALIDATION_CALL = "@validate_call\ndef"
MEZCLA_DIR = gh.form_path(gh.dir_path(__file__), "..")
##
TEST_REGEX = system.getenv_value(
    "TEST_REGEX", None,
    "Regex for tests to include; ex: '^test_c.*' for debugging")
UNDER_UNIX = (os.name == 'posix')
UNDER_UNIX_REASON = "Only applies to Unix"

class TestMisc(TestWrapper):
    """Class for test case definitions"""
    script_module= None

    def get_python_module_files(self, include_tests=False, include_archive=False, include_examples=False):
        """Return list of files for python modules.
        Optionally INCLUDE_TESTS unless configuration module like conftest.py.
        Likewise, optionally INCLUDE_EXAMPLES unless commonly used one like hf_stable_diffusion.
        Also, optionally INCLUDE_ARCHIVE."""
        debug.trace_expr(5, include_tests, include_archive, include_examples,
                    prefix="in get_python_module_files: ")
        ## OLD: all_python_modules = gh.run("find . -name '*.py'").splitlines()
        # Get all filenames for python modules in the repo (i.e., at current commit)
        # Note: only includes python files in mezcla subdir (e.g, mezcla/cut.py
        # and mezcla/tests/test_cut.py but not docs/conf.py).
        debug.assertion(system.get_current_directory().endswith("mezcla"))
        all_python_modules = gh.run(r"git ls-tree -r --name-only HEAD | grep '.*\.py$'").splitlines()
        debug.trace_values(6, all_python_modules)
        debug.assertion(system.intersection(["debug.py", "system.py"], all_python_modules))
        debug.assertion(not system.intersection(["setup.py", "conf.py"], all_python_modules))
        ok_python_modules = []

        # Make sure each module adheres to the basic conventions
        for module in all_python_modules:
            dir_path, filename = system.split_path(module)
            _parent_path, dir_proper = system.split_path(dir_path)
            include = True
            reason = "???"
            if TEST_REGEX and not my_re.search(TEST_REGEX, module):
                include = False
                reason = f"not matching test regex {TEST_REGEX!r}"
            if include and (dir_proper == "tests"):
                include = (include_tests and not filename in ["conftest.py"])
                reason = "test script"
            if include and (dir_proper == "archive"):
                include = include_archive
                reason = "archived script"
            if include and (dir_proper == "examples"):
                INCLUDED_EXAMPLES = ["hf_stable_diffusion.py", "hugging_face_translation.py", "youtube_transcript.py"]
                include = include_examples or (filename in INCLUDED_EXAMPLES)
                reason = "example script (not in {INCLUDED_EXAMPLES!r})"
            if include:
                ok_python_modules.append(module)
            else:
                debug.trace(4, f"FYI: Ignoring module {module!r}: {reason}")
        debug.trace_expr(5, ok_python_modules)
        return ok_python_modules
    
    def transform_for_validation(self, file_path):
        """Creates a temporary copy of the script for validation of argument calls (using pydantic)"""
        debug.trace(6, f"transform_for_validation({file_path})")
        content = system.read_file(file_path)
        content = my_re.sub(r"^def ", r"@validate_call\n\g<0>", content, flags=my_re.MULTILINE)
        ## Uncomment the line below (and comment the line above) if the decorators are previously used
        ## May not be compatible with scripts in mezcla/tests
        ##
        ## OLD:
        ## # content = my_re.sub(r"^(?:(?!\s*[@'#\']).*?)(\s*)(def )", r'\1@validate_call\n\1\2', content, flags=my_re.MULTILINE)
        ## ## TODO1:  something like the following (so that shebang line kept as first):
        ## ## my_re.sub(r"^((from \S+ )?import)", fr"\1{LINE_IMPORT_PYDANTIC}", content, flags=my_re.MULTILINE, count=1)
        ## content = LINE_IMPORT_PYDANTIC + content
        content = my_re.sub(r"^((from \S+ )?import)", fr"{LINE_IMPORT_PYDANTIC}\n\1", content, flags=my_re.MULTILINE, count=1)
        if LINE_IMPORT_PYDANTIC not in content:
            content = LINE_IMPORT_PYDANTIC + content
        ## OLD:
        ## ## TODO2: use self.get_temp_file(): Lorenzo added new functionality
        ## output_path = OUTPUT_PATH_PYDANTIC + gh.basename(file_path)
        output_path = gh.form_path(self.get_temp_dir(), gh.basename(file_path))
        system.write_file(filename=output_path, text=content)
        return output_path

    def save_transformed_for_validation(self, from_dir, to_dir, script_name):
        """Creates a copy of the script for validation of argument calls (using pydantic)"""
        debug.trace_expr(6, from_dir, to_dir, script_name,
                    prefix="in save_transformed_for_validation: ")
        original_path = gh.form_path(from_dir, script_name)
        ## OLD: new_code = self.transform_for_validation(original_path)
        new_code_path = self.transform_for_validation(original_path)
        new_code = system.read_file(new_code_path)
        destination_path = gh.form_path(to_dir, script_name)
        system.write_file(destination_path, new_code)
    
    def run_test(self, label, temp_dir, test_name):
        """Run a test script in a temporary directory.
           Note: Example invocation: run_test("orig", "/tmp/mezcla-original", "html_utils.py")"""
        debug.trace_expr(6, label, temp_dir, test_name,
                    prefix="in run_test: ")
        test_script = gh.form_path(temp_dir, "tests", f"test_{test_name}")
        result = gh.run(f"PYTHONPATH='{temp_dir}/..' pytest {test_script}")
        if debug.verbose_debugging():
            ## OLD:
            ## number = random.randint(10000, 99999)
            ## system.write_file(f"test_{test_name}_{number}.{name}.out", result)
            result_file = gh.form_path(self.get_temp_dir(), f"test_{test_name}.{label}.out")
            system.write_file(result_file, result)
            check_errors_script = gh.run("which check_errors.perl")
            if system.file_exists(check_errors_script):
                gh.run("perl -sw {check_errors_script} {result_file}")
            else:
                debug.trace(5, "FYI: Install shell-script repo for useful utilities like check_errors.perl")
        return result    

    def count_serious_errors(self, results):
        """Return number of errors in RESULTS
        Note: This mostly covers exceptions butalso includes some errors"""
        # This doesn't user check_errors.perl as above because not part of repo
        num_errors = len(my_re.findall(r"Exception|ModuleNotFoundError|RuntimeError|SyntaxError|TypeError", results))
        debug.trace(6, f"count_serious_errors({gh.elide(results)!r}) => {num_errors}")
        return num_errors
    
    def count_failures(self, results):
        """Count the number of failures in the test results"""
        return sum(map(system.to_int, gh.extract_matches_from_text(r"(\d+) x?failed", results)))
    
    @pytest.mark.xfail
    def test_01_check_for_tests(self):
        """Make sure test exists for each Python module"""
        debug.trace(4, "test_01_check_for_tests()")
        for module in self.get_python_module_files(include_tests=False):
            dir_name, filename = system.split_path(module)
            debug.trace_expr(5, module, dir_name, filename)
            include = True
            # Check for cases to skip (e.g., __init__.py)
            if filename.startswith("__"):
                include = False
            # Make sure test file exists
            if include:
                test_path = gh.form_path(dir_name, "tests", f"test_{filename}")
                self.do_assert(system.file_exists(test_path),
                               f"module {module} is missing tests")

    @pytest.mark.xfail
    def test_02_check_lib_usages(self):
        """Make sure modules don't use certain standard library calls directly
        Note: The mezcla equivalent should be used for sake of debugging traces"""
        debug.trace(4, "test_02_check_lib_usages()")
        for module in self.get_python_module_files(include_tests=False):
            module_code = system.read_file(module)
            bad_usage = False
            # Direct use of sys.stdin.read()
            # TODO3: use bash AST (e.g., to exclude comments)
            if "import sys" in module_code and "sys.stdin.read" in module_code:
                bad_usage = (module != "main.py")
                debug.trace(4, f"Direct use of sys.stdin.read in {module}")
            self.do_assert(not bad_usage, f"module {module} has bad sys.stdin.read usage")

    @pytest.mark.xfail
    def test_03_check_mezcla_convention(self):
        """Make sure modules follow various mezcla conventions
        - file execute permission, 
        - no tabs"""
        ## TODO3: python-lint alias produces no warning
        debug.trace(4, "test_03_check_permissions()")
        for module in self.get_python_module_files(include_tests=True):
            has_execute_perm = gh.run(f'ls -l "{module}" | grep ^...x..x..x')
            self.do_assert(has_execute_perm, f"module {module} is missing execute permission")
            has_tab_char = "\t" in system.read_file(module)
            self.do_assert(not has_tab_char, f"module {module} should include tab chars")

    @pytest.mark.skipif(SKIP_SLOW_TESTS, reason=SKIP_SLOW_REASON)
    @pytest.mark.xfail
    def test_04_check_transform_for_validation(self):
        """Make sure the transformation is successful when adding decorators for validation of function calls"""
        ## TODO2: have option to run scripts to check for runtime errors
        debug.trace(4, "test_04_check_transform_for_validation()")
        ## OLD: TMP_DIR = system.getenv("TMP")
        for module in self.get_python_module_files(include_tests=False):
            ## OLD:
            ## TODO2: use gh.resolve_path
            ## input_file = "../html_utils.py"
            input_file = module
            result = self.transform_for_validation(input_file)
            ## OLD:
            ## ## TODO1: don't hardcode refs to /tmp 
            ## assert "temp" in result and "/tmp/" in result
            ## OLD: assert "temp" in result and TMP_DIR in result
            content = system.read_file(result)
            ## For debugging: print(content)
            ## OLD: assert content.startswith(LINE_IMPORT_PYDANTIC)
            assert LINE_IMPORT_PYDANTIC in content
            ## OLD: assert content.count(DECORATOR_VALIDATION_CALL) >= 1
            old_content = system.read_file(input_file)
            if "\ndef " in old_content:
                assert content.count(DECORATOR_VALIDATION_CALL) >= 1
            else:
                debug.trace(4, f"FYI: Ignoring validation call check for {input_file}, which has no def's")

    @pytest.mark.skipif(SKIP_SLOW_TESTS, reason=SKIP_SLOW_REASON)
    @pytest.mark.skipif(not UNDER_UNIX, reason=UNDER_UNIX_REASON)
    @pytest.mark.xfail
    def test_05_usage_statements(self):
        """Make sure usage statments refer to valid arguments"""
        debug.trace(4, "test_05_usage_statements()")
        repo_dir = gh.run("git rev-parse --show-toplevel")
        timeout_script = gh.form_path(repo_dir, "tools", "cmd.sh")
        assert system.file_exists(timeout_script)
        
        # note: checks for incomplete change like --include-header => --header in randomize_lines.py
        #   usage: randomize_lines.py [-h] [--header] [--verbose] [--seed SEED] [--percent PERCENT] [filename]
        #   Sample usage:
        #      randomize_lines.py --include-header --percent 10 ./examples/pima-indians-diabetes.csv
        for module in self.get_python_module_files(include_tests=False):
            debug.trace(4, f"Checking {module} usage")
            module_usage = gh.run(f"echo '' | {timeout_script} {module} --help")
            module_synopsis = my_re.search(r"^\s*usage:.*", module_usage)
            # note: use non-greedy search for perl-like paragraph search
            for usage in my_re.findall(r"^ *(advanced|example|s[ai]mple|typical) +usage:.*+\n\n",
                                       module_usage, flags=my_re.MULTILINE|my_re.DOTALL):
                for option in my_re.findall(r"--[\w-]+", usage):
                    self.do_assert(option in module_synopsis)

    @pytest.mark.skipif(SKIP_SLOW_TESTS, reason="Skipping long tests")
    @pytest.mark.xfail
    def test_06_type_hinting(self):
        """Test type hinting by running tests over code with pydantic decorators.
        Note: This creates copy of mezcla scripts and checks pytest result over original vs modified.
        Warning: TEST_REGEX must be set to *.py to run all scripts; otherwise main scripts tested(e.g., system.py, glue_helpers.py, etc.).
        """
        debug.trace(4, "test_06_type_hinting()")
        ## TODO: create a module with very broken python code to test the test (for false positives)
        main_scripts = [
            "debug.py", "glue_helpers.py", "html_utils.py",
            "my_regex.py", "system.py", "unittest_wrapper.py",
        ]
        orig_mezcla_dir = MEZCLA_DIR
        ## OLD: temp_mezcla_dir = self.temp_file + "-mezcla"
        temp_mezcla_dir = gh.form_path(self.get_temp_dir(), "mezcla")

        # Create copy of mezcla scripts (n.b., expedient to allow simple PYTHONPATH change)
        temp_mezcla_test_dir = gh.form_path(temp_mezcla_dir, "tests")
        ## TODO2: copy_dir
        gh.full_mkdir(temp_mezcla_test_dir)
        gh.run(f"cp -vf {orig_mezcla_dir}/*.py {temp_mezcla_dir}")
        gh.run(f"cp -vf {orig_mezcla_dir}/tests/*.py {temp_mezcla_test_dir}")

        # Test each of the main scripts and check for test result differences
        num_bad = 0
        test_scripts = (main_scripts if not TEST_REGEX
                        else self.get_python_module_files(include_tests=False))
        num_cases = len(main_scripts)
        for script in main_scripts:
            # Add dynamic type checking to scripts (e.g., via pydantic)
            self.save_transformed_for_validation(orig_mezcla_dir, temp_mezcla_dir, script)

            # Run tests using transformed and original script
            temp_results = self.run_test("temp", temp_mezcla_dir, script)
            orig_results = self.run_test("orig", orig_mezcla_dir, script)

            # Check for errors
            # exs: 1) more exceptions in transformed run vs original; 2) differences in number of failures
            bad_results = False
            ## OLD: if (temp_results.count("Exception") > orig_results.count("Exception")):
            if (self.count_serious_errors(temp_results) > self.count_serious_errors(orig_results)):
                debug.trace(4, f"Warning: more exceptions or errors running test of transformed script {script}")
                bad_results = True
            num_temp_failed = self.count_failures(temp_results)
            num_orig_failed = self.count_failures(orig_results)
            debug.assertion(num_temp_failed >= num_orig_failed)
            debug.trace_expr(5, num_temp_failed, num_orig_failed, script)
            if (num_temp_failed > num_orig_failed):
                bad_results = True
            if bad_results:
                num_bad += 1

        # Only allow for a relatively small number of failures
        bad_pct = round(num_bad / num_cases * 100, 2)
        debug.trace_expr(5, bad_pct, num_bad, num_cases)
        assert bad_pct < 20

                    
#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Tests for format_profile module
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_train_language_model.py
#
#-------------------------------------------------------------------------------
# Tested script usage:
#
#   $ format_profile.py 
#   
#   Usage: /home/tomohara/Mezcla/mezcla/format_profile.py profile-log
#   
#   Notes:
#   - Use FULL_PATH to include directoy for filename
#     (e.g., helps to resolve all those __init__.py entries).
#   - Use PROFILE_KEY to over default sorting (cumulative).
#   - Main keys: 
#          cumtime, filename, ncalls, tottime
#   - Other keys: 
#          module, pcalls, line, name, nfl, stdname
#   - Alternative keys:
#          calls, cumulative, file, time
#   - Unfortunately, memory profiling is not supported.
#   - For more details, check following:
#       http://docs.python.org/3/library/profile.html
#   
#   Example (assumes bash):
#       $ python -m cProfile -o /tmp/profile.data simple_main_example.py
#       $ PROFILE_KEY=calls /home/tomohara/Mezcla/mezcla/format_profile.py /tmp/profile.data | head
#
#-------------------------------------------------------------------------------
# Sample tested output:
#
#   Fri Oct  3 21:58:02 2025    /tmp/test_format_profile/test-1-0
#   
#            48514 function calls (47753 primitive calls) in 0.029 seconds
#   
#      Ordered by: call count
#   
#      ncalls  tottime  percall  cumtime  percall filename:lineno(function)
#        4212    0.000    0.000    0.000    0.000 {built-in method builtins.isinstance}
#        3887    0.000    0.000    0.000    0.000 {method 'rstrip' of 'str' objects}
#   2057/1982    0.000    0.000    0.000    0.000 {built-in method builtins.len}
#        2013    0.000    0.000    0.000    0.000 {method 'join' of 'str' objects}
#        1957    0.000    0.000    0.000    0.000 <frozen importlib._bootstrap>:244(_verbose_message)
#        1897    0.001    0.000    0.002    0.000 <frozen importlib._bootstrap_external>:126(_path_join)
#        1897    0.001    0.000    0.001    0.000 <frozen importlib._bootstrap_external>:128(<listcomp>)
#        1011    0.000    0.000    0.000    0.000 {method 'append' of 'list' objects}
#         792    0.000    0.000    0.000    0.000 {built-in method builtins.hasattr}
#         790    0.000    0.000    0.000    0.000 {built-in method builtins.getattr}
#
#-------------------------------------------------------------------------------
## TODO0: convert test_formatprofile_PK_cumtime through test_formatprofile_PK_tottime using
##    an AI assistant such as Clause Opus 4.1! Follow Tom's tips on getting tests to be more mezcla
##    like.
## TODO1: Very carefully review test/template.py,
## TODO2: Follow the mezcla conventions! For example, do not comment out code without
##        indicating why! See previous todo.
##
## TIP:
## - Only need to test a few main keys in any depth, so generalize test_formatprofile_PK_calls
##   so that it can handle alternative sort orders without substantial revision.
## - The minor keys can just check for "Ordered by" label differences.
##

"""Tests for format_profile module"""

# Standard packages
from typing import Optional

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                        global module object
#    TestIt.script_module:              path to file
import mezcla.format_profile as THE_MODULE

class TestFormatProfile(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    use_temp_base_dir = True    # treat TEMP_BASE as directory
    # note: The run involves a usage statement from an archived script to minimize maintenance.
    # Aside: This also makes the test much quicker than the previous version using test_glue_helpers.
    testing_env = "USE_LUCENE=0"
    testing_script = gh.resolve_path(gh.form_path("archive", "search_table_file_index.py"),
                                     heuristic=True)
    testing_options = ""
    old_testing_script = gh.resolve_path(gh.form_path("test_glue_helpers.py"), heuristic=True)

    @staticmethod
    def encode(output, regex=False):
        r"""Convert OUTPUT to encoded representation (e.g., N for each digit).
        Optionally, encodes as REGEX and replaces "..." with ".*".
        For example: 
        >>> TestFormatProfile.encode('0.077 debug.py\n...0.052 system.py', regex=True)
        'N\.NNN debug\.py.*N\.NNN system\.py'
        """
        # EX: TestFormatProfile.encode("7    0.000    0.000    0.540    0.077 debug.py:608(trace_expr)") => "N N.NNN N.NNN N.NNN N.NNN debug.py:NNN(trace_expr)
        in_output = output.strip()
        output = my_re.sub(r"[ \t]+", " ", output)
        output = my_re.sub(r"^ *", "", output, flags=my_re.MULTILINE)
        output = my_re.sub(r" \n", " ", output, flags=my_re.MULTILINE)
        output = my_re.sub(r"[0-9]", "N", output)
        if regex:
            output = my_re.escape(output)
            output = output.replace(r"\.\.\.", ".*")
        debug.trace(7, f"encode({in_output!r}) => {output!r}")
        return output
    
    def helper_format_profile(
            self, key_arg, order_indicator: Optional[str] = None,
            good_sample_output: Optional[str] = None, bad_sample_output: Optional[str] = None,
            testing_script: Optional[str] = None, testing_env: Optional[str] = None, testing_options: Optional[str] = None):
        """Helper function for format_profile tests, using KEY_ARG to order and verifying via ORDER_INDICATOR.
        The optional arguments give GOOD_SAMPLE_OUTPUT and BAD_SAMPLE_OUTPUT. Each should be a multiline string
        in order to test for proper ordering of entries. 
        Note: Digits will be encoded by N to make the tests more robust. In addition, whitespace will be collapsed.
        Other optional arguments give the TESTING_SCRIPT, TESTING_ENV, and TESTING_OPTIONS.

        Sample invocation: 
            GOOD_SAMPLE_OUTPUT = "    11    0.000    0.000    0.625    0.057 __init__.py:1(<module>)\n    147/1    0.001    0.000    0.583    0.583 {built-in method builtins.exec}"
            BAD_SAMPLE_OUTPUT = "        2    0.000    0.000    0.254    0.127 asttokens.py:132(mark_tokens)\n        1    0.000    0.000    0.340    0.340 debug.py:1366(debug_init)"
            output = self.helper_format_profile("cumtime", "cumulative time", GOOD_SAMPLE_OUTPUT, BAD_SAMPLE_OUTPUT)
        """

        # Get options
        if testing_script is None:
            testing_script = self.testing_script
        if testing_env is None:
            testing_env = (self.testing_env if (testing_script == self.testing_script) else "")
        if testing_options is None:
            testing_options = self.testing_options
        debug.assertion(system.file_exists(testing_script))

        # Run script to produce profile and then run profile formatter
        # note: It runs under the default debugging level to avoid extraenous profiling entries.
        profile_log = self.get_temp_file()
        command_cprofile = f"python3 -m cProfile -o {profile_log} {testing_script}"
        _cprofile_output = gh.run(f"{testing_env} {command_cprofile} {testing_options}")
        output = self.run_script(
            env_options=f"DEBUG_LEVEL=2 PROFILE_KEY={key_arg}",
            data_file=profile_log,
        )

        # Make sure ordered by desired metric
        assert("Ordered by:" in output)
        if order_indicator:
            assert(my_re.search(fr"Ordered by: *{order_indicator}", output))
        encoded_output = self.encode(output)
        debug.assertion("N.NNN" in encoded_output)
        if good_sample_output:
            assert(my_re.search(self.encode(good_sample_output, regex=True), encoded_output,
                                flags=my_re.DOTALL|my_re.MULTILINE))
        if bad_sample_output:
            assert(not my_re.search(self.encode(bad_sample_output, regex=True), encoded_output,
                                    flags=my_re.DOTALL|my_re.MULTILINE))

        return output

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_calls(self):
        """Ensures that PROFILE_KEY=calls works as expected"""
        debug.trace(4, f"test_formatprofile_PK_calls(); self={self}")

        # Specify good and bad output:
        # ncalls tottime  percall  cumtime  percall filename:lineno(function)
        GOOD_SAMPLE_OUTPUT = (
            "1    0.000    0.000    0.000    0.000 search_table_file_index.py:545(main)\n" +
            "..." + 
            "1    0.000    0.000    0.573    0.573 search_table_file_index.py:1(<module>)\n")
        BAD_SAMPLE_OUTPUT = (
            "14986    0.021    0.000    0.041    0.000 tokenize.py:433(_tokenize)\n" +
            "..." +            
            "15087    0.002    0.000    0.003    0.000 inspect.py:283(ismodule)\n")
        debug.trace(4, f"helper_format_profile(); self={self}")
        _output = self.helper_format_profile("calls", "call count", GOOD_SAMPLE_OUTPUT, BAD_SAMPLE_OUTPUT)
        return
        
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_cumulative(self):
        """Ensures that PROFILE_KEY=cumulative works as expected"""

        GOOD_SAMPLE_OUTPUT = "<frozen importlib._bootstrap_externals>:877(exec_module)"
        BAD_SAMPLE_OUTPUT = "1    0.000    0.000    0.000    0.000 {built-in method posix.readlink}"

        debug.trace(4, f"test_formatprofile_PK_cumulative(); self={self}")
        _output = self.helper_format_profile("cumulative", "cumulative time", GOOD_SAMPLE_OUTPUT, BAD_SAMPLE_OUTPUT)
        return
    
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_cumtime(self):
        """Ensures that PROFILE_KEY=cumtime works as expected"""
        ## TODO: Find other input sample
        debug.trace(4, f"test_formatprofile_PK_cumtime(); self={self}")
        key_arg = "cumtime"
        SAMPLE_OUTPUT = [
            "test_glue_helper.py:1(<module>)", # Incorrect Line 
            "1    0.000    0.000    0.000    0.000 {method 'fileno' of '_io.BufferedWriter' objects}"
            ]

        # output = gh.read_file(empty_file1)
        output = self.helper_format_profile(key_arg, testing_script=self.old_testing_script)
        # print(output)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        # return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_file(self):
        """Ensures that PROFILE_KEY=file works as expected"""

        key_arg = "file"
        SAMPLE_OUTPUT = [
            "<frozen importlib._bootstraps>:391(cached)", 
            "1    0.000    0.000    0.000    0.000 {method 'union' of 'frozenset' objects}"
            ]
        
        debug.trace(4, f"test_formatprofile_PK_file(); self={self}")
        output = self.helper_format_profile(key_arg, testing_script=self.old_testing_script)

        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
    
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_filename(self):
        """Ensures that PROFILE_KEY=filename works as expected"""

        key_arg = "filename"
        SAMPLE_OUTPUT = [
            "1    0.000    0.000    0.000    0.000 :1(ReprEntryNativeAttributes)", 
            "6768    0.000    0.000    0.000    0.000 :1(ReprEntryAttributes)"
        ]


        debug.trace(4, f"test_formatprofile_PK_filename(); self={self}")
        output = self.helper_format_profile(key_arg, testing_script=self.old_testing_script)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] not in output)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_module(self):
        """Ensures that PROFILE_KEY=module works as expected"""

        key_arg = "module"
        SAMPLE_OUTPUT = [
            "ElementTree.pytest:1771(C14NWriterTarget)",
            "1    0.000    0.000    0.000    0.000 terminal.py:1306(_build_normal_summary_stats_line)"
        ]

        debug.trace(4, f"test_formatprofile_PK_module(); self={self}")
        output = self.helper_format_profile(key_arg, testing_script=self.old_testing_script)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_ncalls(self):
        """Ensures that PROFILE_KEY=ncalls works as expected"""
        ## TODO: Find other input sample

        key_arg = "ncalls"
        ## OLD: testing_script = old_testing_script
        SAMPLE_OUTPUT = [
            "{method 'extend' of 'collections.deque' objections}", 
            "1    0.000    0.000    0.000    0.000 test_glue_helpers.py:385(test_get_files_matching_specs)"
            ]

        debug.trace(4, f"test_formatprofile_PK_ncalls(); self={self}")

        output = self.helper_format_profile(key_arg, testing_script=self.old_testing_script)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_pcalls(self):
        """Ensures that PROFILE_KEY=pcalls works as expected"""

        key_arg = "pcalls"
        SAMPLE_OUTPUT = [
            "unix_events.py:1022(SafestChildWatcher)",
            "1    0.000    0.000    0.000    0.000 <frozen importlib._bootstrap>:294(_module_repr)"
        ]

        debug.trace(4, f"test_formatprofile_PK_pcalls(); self={self}")
        output = self.helper_format_profile(key_arg, testing_script=self.old_testing_script)
        
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        # return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_line(self):
        """Ensures that PROFILE_KEY=line works as expected"""

        key_arg = "line"
        SAMPLE_OUTPUT = [
            "{method 'with_traceback' of 'BaseExceptions' objects}", 
            "1    0.000    0.000    0.000    0.000 {method 'rjust' of 'str' objects}"
        ]
        debug.trace(4, f"test_formatprofile_PK_line(); self={self}")

        output = self.helper_format_profile(key_arg, testing_script=self.old_testing_script)

        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_name(self):
        """Ensures that PROFILE_KEY=name works as expected"""

        key_arg = "name"
        SAMPLE_OUTPUT = [
            "{method 'fileno' of '_io.BufferedReaders' objects}", 
            "6    0.000    0.000    0.000    0.000 {method 'pop' of 'collections.deque' objects}"
        ]

        debug.trace(4, f"test_formatprofile_PK_name(); self={self}")
        output = self.helper_format_profile(key_arg, testing_script=self.old_testing_script)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_nfl(self):
        """Ensures that PROFILE_KEY=nfl works as expected"""

        key_arg = "nfl"
        
        SAMPLE_OUTPUT = [
            "{built-in methods _imp.is_frozen}",
            "1    0.000    0.000    0.000    0.000 {built-in method _stat.S_IMODE}", 
        ]

        debug.trace(4, f"test_formatprofile_PK_nfl(); self={self}")
        
        output = self.helper_format_profile(key_arg, testing_script=self.old_testing_script)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_stdname(self):
        """Ensures that PROFILE_KEY=stdname works as expected"""

        key_arg = "stdname"

        SAMPLE_OUTPUT = [
            "zipperfile.py:1(<module>)", 
            "1    0.000    0.000    0.000    0.000 _synchronization.py:70(SemaphoreStatistics)"
        ]

        debug.trace(4, f"test_formatprofile_PK_stdname(); self={self}")

        output = self.helper_format_profile(key_arg, self.testing_script)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_time(self):
        """Ensures that PROFILE_KEY=time works as expected"""

        key_arg = "time"

        SAMPLE_OUTPUT = [
            "<frozen importlibrary._bootstrap_external>:380(cache_from_source)", 
            "1    0.000    0.000    0.000    0.000 glue_helpers.py:731(delete_existing_file)"
        ]

        debug.trace(4, f"test_formatprofile_PK_time(); self={self}")
        output = self.helper_format_profile(key_arg, testing_script=self.old_testing_script)
        
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_tottime(self):
        """Ensures that PROFILE_KEY=tottime works as expected"""
        ## TODO: Find other input sample
        
        key_arg = "tottime"
        SAMPLE_OUTPUT = [
            "{method 'split' of 're.Pattern' objections}", 
            "1    0.000    0.000    0.000    0.000 cacheprovider.py:390(pytest_sessionfinish)"
        ]

        debug.trace(4, f"test_formatprofile_PK_tottime(); self={self}")
        
        output = self.helper_format_profile(key_arg, testing_script=self.old_testing_script)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        return

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Test(s) for ../misc_utils.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_misc_utils.py
#

"""Tests for misc_utils module"""

# Standard packages
import math
import datetime
import time
import os
from typing import List

# Installed packages
import pytest
from unittest_parametrize import ParametrizedTestCase, parametrize

# Local packages
from mezcla import glue_helpers as gh
from mezcla import debug
from mezcla.my_regex import my_re
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla.tests.common_module import SKIP_EXPECTED_ERRORS, SKIP_EXPECTED_REASON
## OLD: from mezcla.mezcla_to_standard import EqCall, Features
# note: mezcla_to_standard uses packages not installed by default (e.g., libcst)
try:
    from mezcla import mezcla_to_standard
    EqCall = mezcla_to_standard.EqCall
except:
    mezcla_to_standard = None
    class Path:
        """Dummy path"""
        path = "n/a"
    class EqCall:
        """Dummy equivalent call"""
        ## TODO3: use mock package
        dests = targets = [Path()]
        def __init__(self, *_args, **_kwargs):
            pass
    debug.trace_exception(4, "mezcla.mezcla_to_standard import")

# Note: Two references are used for the module to be tested:
#    THE_MODULE:            global module object
import mezcla.misc_utils as THE_MODULE

# Make sure more_itertools available
more_itertools = None
try:
    import more_itertools
except ImportError:
    system.print_exception_info("more_itertools import")

# Constants
ONE_MB = 1024 ** 2


class TestMiscUtils(TestWrapper, ParametrizedTestCase):
    """Class for test case definitions"""
    # note: script_module used in argument parsing sanity check (e.g., --help)
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    def test_transitive_closure(self):
        """Ensure transitive_closure works as expected"""
        debug.trace(4, "test_transitive_closure()")

        actual = THE_MODULE.transitive_closure([(1, 2), (2, 3), (3, 4)])
        expected = set([(1, 2), (1, 3), (1, 4), (2, 3), (3, 4), (2, 4)])
        assert actual == expected

    def test_read_tabular_data(self):
        """Ensure read_tabular_data works as expected"""
        debug.trace(4, "test_read_tabular_data()")
        string_table = (
            'language\tPython\n' +
            'framework\tPytest\n'
        )
        dict_table = {
            'language': 'Python\n',
            'framework': 'Pytest\n',
        }
        temp_file = self.get_temp_file()
        system.write_file(temp_file, string_table)
        assert THE_MODULE.read_tabular_data(temp_file) == dict_table

    def test_extract_string_list(self):
        """Ensure extract_string_list works as expected"""
        debug.trace(4, "test_extract_string_list()")

        # Check typical cases
        assert THE_MODULE.extract_string_list("  a  b,c") == ['a', 'b', 'c']
        assert THE_MODULE.extract_string_list("a\nb\tc") == ['a', 'b', 'c']
        assert THE_MODULE.extract_string_list("a, b c") == ['a', 'b', 'c']

        # Check special cases
        assert THE_MODULE.extract_string_list("") == []
        assert THE_MODULE.extract_string_list(",") == ['', '']
        assert THE_MODULE.extract_string_list(",", strip_empty=True) == []

    def test_extract_numeric_string_list(self):
        """Verify extract_string_list with numericranges"""
        debug.trace(4, "test_extract_numeric_string_list()")

        # Check typical cases
        self.do_assert(THE_MODULE.extract_string_list("1-4", allow_numeric_ranges=True)
                       == ['1', '2', '3', '4'])
        self.do_assert(THE_MODULE.extract_string_list("a b-c d", allow_numeric_ranges=True)
                       == ['a', 'b-c', 'd'])
        
    def test_is_prime(self):
        """Ensure is_prime works as expected"""
        debug.trace(4, "test_is_prime()")

        first_100_primes = [
            2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37,
            41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83,
            89, 97, 101, 103, 107, 109, 113, 127, 131,
            137, 139, 149, 151, 157, 163, 167, 173, 179,
            181, 191, 193, 197, 199, 211, 223, 227, 229,
            233, 239, 241, 251, 257, 263, 269, 271, 277,
            281, 283, 293, 307, 311, 313, 317, 331, 337,
            347, 349, 353, 359, 367, 373, 379, 383, 389,
            397, 401, 409, 419, 421, 431, 433, 439, 443,
            449, 457, 461, 463, 467, 479, 487, 491, 499,
            503, 509, 521, 523, 541,
        ]

        assert all(THE_MODULE.is_prime(n) for n in first_100_primes)
        assert all((not THE_MODULE.is_prime(n)) for n in range(first_100_primes[-1]) if n not in first_100_primes)

    def test_fibonacci(self):
        """Ensure fibonacci works as expected"""
        debug.trace(4, "test_fibonacci()")
        assert not THE_MODULE.fibonacci(-5)
        assert THE_MODULE.fibonacci(1) == [0]
        assert THE_MODULE.fibonacci(10) == [0, 1, 1, 2, 3, 5, 8]

    def test_sort_weighted_hash(self):
        """Ensure sort_weighted_hash works as expected"""
        debug.trace(4, "test_sort_weighted_hash()")
        test_hash = {
            'bananas': 3,
            'apples': 1411,
            'peach': 43,
        }
        sorted_hash = [
            ('bananas', 3),
            ('peach', 43),
            ('apples', 1411),
        ]
        reversed_hash = [
            ('apples', 1411),
            ('peach', 43),
            ('bananas', 3),
        ]
        assert THE_MODULE.sort_weighted_hash(test_hash) == reversed_hash
        assert THE_MODULE.sort_weighted_hash(test_hash, reverse=False) == sorted_hash
        assert len(THE_MODULE.sort_weighted_hash(test_hash, max_num=2)) == 2

    def test_unzip(self):
        """Ensure unzip works as expected"""
        debug.trace(4, "test_unzip()")
        assert THE_MODULE.unzip(zip([1, 2, 3], ['a', 'b', 'c'])) == [[1, 2, 3], ['a', 'b', 'c']]
        assert THE_MODULE.unzip(zip([], []), 2) == [[], []]
        assert THE_MODULE.unzip(zip(), 4) == [[], [], [], []]
        assert THE_MODULE.unzip(zip(), 4) != [[], []]

    def test_get_current_frame(self):
        """Ensure get_current_frame works as expected"""
        debug.trace(4, "test_get_current_frame()")
        stack = str(THE_MODULE.get_current_frame())
        test_name = system.get_current_function_name()
        assert f'code {test_name}' in stack
        assert repr(__file__) in stack

    def test_eval_expression(self):
        """Ensure eval_expression works as expected"""
        debug.trace(4, "test_eval_expression()")
        assert THE_MODULE.eval_expression("len([123, 321]) == 2")
        assert not THE_MODULE.eval_expression("'helloworld' == 2")

    def test_trace_named_object(self):
        """Ensure trace_named_object works as expected"""
        debug.trace(4, "test_trace_named_object()")
        # With level -1 we ensure that the trace will be printed
        THE_MODULE.trace_named_object(-1, "sys.argv")
        captured = self.get_stderr()
        assert "sys.argv" in captured

    def test_trace_named_objects(self):
        """Ensure trace_named_objects works as expected"""
        debug.trace(4, "test_trace_named_objects()")
        # With level -1 we ensure that the trace will be printed
        THE_MODULE.trace_named_objects(-1, "[len(sys.argv), sys.argv]")
        captured = self.get_stderr()
        assert "len(sys.argv)" in captured
        assert "sys.argv" in captured

    @pytest.mark.skipif(not more_itertools, reason="Unable to load more_itertools")
    def test_exactly1(self):
        """Ensure exactly1 works as expected"""
        debug.trace(4, "test_exactly1()")
        assert THE_MODULE.exactly1([False, False, True])
        assert not THE_MODULE.exactly1([False, True, True])
        assert not THE_MODULE.exactly1([False, False, False])
        assert not THE_MODULE.exactly1([])

    def test_string_diff(self):
        """Ensure string_diff works as expected"""
        debug.trace(4, "test_string_diff()")

        string_one = 'one\ntwo\nthree\nfour'
        string_two = 'one\ntoo\ntree\nfour'
        expected_diff = (
            '  one\n'
            '< two\n'
            '…  ^\n'
            '> too\n'
            '…  ^\n'
            '< three\n'
            '…  -\n'
            '> tree\n'
            '  four\n'
        )

        assert THE_MODULE.string_diff(string_one, string_two) == expected_diff

    def test_elide_string_values(self):
        """Ensure elide_string_values works as expected"""
        debug.trace(4, "test_elide_string_values()")
        hello = "hello"
        assert 'hell...' == THE_MODULE.elide_string_values(hello, max_len=4)

    def test_is_close(self):
        """ensure is_close works as expected"""
        debug.trace(4, "test_is_close()")
        assert     THE_MODULE.is_close(1.0 + 0.999 , 2.0, 0.001)
        assert not THE_MODULE.is_close(1.0 + 0.999 , 2.0, 0.0001)

    def test_get_date_ddmmmyy(self):
        """ensure get_date_ddmmmyy works as expected"""
        debug.trace(4, "test_get_date_ddmmmyy()")
        assert THE_MODULE.get_date_ddmmmyy(datetime.date(2004, 9, 16)) == '16sep04'

    def test_parse_timestamp(self):
        """ensure parse_timestamp works as expected"""
        debug.trace(4, "test_parse_timestamp()")
        ts = datetime.datetime(2004, 9, 16, 12, 30, 25, 123123)
        ts_iso = '2004-09-16T12:30:25.1231234Z'
        ts_iso_2 = '2004-09-16T12:30:25.123123Z'
        parsed_ts = THE_MODULE.parse_timestamp(ts_iso, truncate=True)
        parsed_ts_truncated = THE_MODULE.parse_timestamp(ts_iso_2, truncate=False)
        assert parsed_ts == ts
        assert parsed_ts_truncated == ts

    def test_add_timestamp_diff(self):
        """ensure add_timestamp_diff works as expected"""
        debug.trace(4, "test_add_timestamp_diff()")
        timestamp = "timestamp 2004-09-16T12:30:25.1231234Z"
        file_in = f"{self.temp_file}.in"
        file_out = f"{self.temp_file}.out"
        system.write_file(file_in, timestamp)
        THE_MODULE.add_timestamp_diff(file_in, file_out)
        contents = system.read_file(file_out)
        assert contents == f"{timestamp} [0]\n"

    def test_random_int(self):
        """ensure random_int works as expected"""
        debug.trace(4, "test_random_int()")
        assert 0 <= THE_MODULE.random_int(0,4) <= 4

    def test_random_float(self):
        """ensure random_float works as expected"""
        debug.trace(4, "test_random_float()")
        assert 0 <= THE_MODULE.random_float(0,4.3) < 4.3

    def test_time_function(self):
        """ensure time_function works as expected"""
        debug.trace(4, "test_time_function()")
        ms = THE_MODULE.time_function(time.sleep, 0.25)
        assert math.floor(ms) == 250

    def test_get_class_from_name(self):
        """ensure get_class_from_name works as expected"""
        debug.trace(4, "test_get_class_from_name()")
        result_class = THE_MODULE.get_class_from_name('date', 'datetime')
        assert result_class is datetime.date

    def check_apply_numeric_suffixes(self, text, expect, just_once=False):
        """Helper for test_apply_numeric_suffixes """
        debug.trace_expr(4, text, expect, prefix="in check_apply_numeric_suffixes: ")
        actual = THE_MODULE.apply_numeric_suffixes(text, just_once=just_once)
        assert actual.strip() == expect.strip()
    
    @pytest.mark.xfail                  # TODO: remove xfail
    @parametrize(
        "text, actual",
        ## TODO? ("text, actual")     # see https://github.com/adamchainz/unittest-parametrize
        [ (f"{ONE_MB - 1024 - 1} {ONE_MB} {ONE_MB + 1024}", "1022.999K 1M 1.001M"),
          ("0000", "0K"),                         # special cas
          ## TODO2: ("-00000000-", "-_00000000_?-"),
          ("-00000000-", "-0K-"),                 # esnure nuemric context
          (f"{int(1.5 * ONE_MB)}", "1.5M"),       # make sure not rounded
         ])
    def test_apply_numeric_suffixes(self, text, actual):
        """Check apply_numeric_suffixes"""
        debug.trace(5, f"TestIt.test_apply_numeric_suffixes(); self={self}")
        self.check_apply_numeric_suffixes(text, actual)

    def test_simple_apply_numeric_suffixes(self):
        """Check apply_numeric_suffixes"""
        debug.trace(5, f"TestIt.test_simple_apply_numeric_suffixes(); self={self}")
        self.check_apply_numeric_suffixes("2048 16384", "2K 16K")
        self.check_apply_numeric_suffixes("2048 16384", "2K 16384", just_once=True)

    @pytest.mark.xfail                  # TODO: remove xfail
    def test_stdin_apply_numeric_suffixes(self):
        """Check apply_numeric_suffixes_stdin"""
        debug.trace(5, f"TestIt.test_stdin_apply_numeric_suffixes(); self={self}")
        temp_file = self.create_temp_file("999 1024")
        self.monkeypatch.setattr("sys.stdin", system.open_file(temp_file))
        THE_MODULE.apply_numeric_suffixes_stdin()
        actual = self.get_stdout().strip()
        assert actual == "999 1K"

    @pytest.mark.xfail                  # TODO: remove xfail
    def test_GlobalSetter(self):
        """Ensure that GlobalSetter's changes are temporary"""
        self.patch_trace_level(3)
        unique_mod_3 = unique_mod_4 = []
        def get_unique_moduli(n, l):
            """Returns unique modulus results over N in [0, l-1]"""
            return system.unique_items(i % n for i in range(l))
        unique_mod_4 = get_unique_moduli(4, 13)
        with THE_MODULE.GlobalSetter(debug, 'trace_level', 8):
            unique_mod_3 = get_unique_moduli(3, 13)
        stderr = self.get_stderr()
        assert my_re.search("unique_items.*0, 1, 2", stderr)
        assert not my_re.search("unique_items.*0, 1, 2, 3", stderr)
        assert system.difference(unique_mod_4, unique_mod_3) == [3]

@pytest.mark.skipif(not mezcla_to_standard, reason="Unable to load mezcla_to_standard")
class TestFileToInstance(TestWrapper):
    """Class for testing convert_file_to_instances: external file loading to support mezcla_to_standard"""
    # note: script_module used in argument parsing sanity check (e.g., --help)
    ## TODO3: Add parameters to reduce redundancy in the various test_convert_*_to_instance methods,
    ## such as resource data file and conversion function.
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    expected_instances = [
        EqCall(gh.rename_file, dests=os.rename),
        EqCall(gh.dir_path, dests=os.path.dirname, eq_params={"filename": "p"}),
    ]

    @pytest.mark.xfail                  # TODO: remove xfail
    def check_instances(self, instances):
        """Check that INSTANCES match expected_instances"""
        assert self.expected_instances[0].targets[0].path == instances[0].targets[0].path
        assert self.expected_instances[0].dests[0].path == instances[0].dests[0].path

        assert self.expected_instances[1].targets[0].path == instances[1].targets[0].path
        assert self.expected_instances[1].dests[0].path == instances[1].dests[0].path
        
    
    @pytest.mark.xfail                  # TODO: remove xfail
    def test_convert_json_to_instance(self):
        """ensure convert_json_to_instance works as expected"""
        debug.trace(4, "test_convert_json_to_instance()")

        json_data = gh.form_path(gh.dirname(__file__), "resources", "instances.json")

        instances: List[EqCall] = THE_MODULE.convert_json_to_instance(
            json_data,
            "mezcla.mezcla_to_standard",
            "EqCall",
            mezcla_to_standard.EQCALL_FIELDS,
        )
        self.check_instances(instances)
        
    @pytest.mark.xfail                  # TODO: remove xfail
    def test_convert_yaml_to_instance(self):
        """ensure convert_yaml_to_instance works as expected"""
        debug.trace(4, "test_convert_yaml_to_instance()")

        yaml_data = gh.form_path(gh.dirname(__file__), "resources", "instances.yaml")

        instances: List[EqCall] = THE_MODULE.convert_yaml_to_instance(
            yaml_data,
            "mezcla.mezcla_to_standard",
            "EqCall",
            mezcla_to_standard.EQCALL_FIELDS,
        )
        self.check_instances(instances)

    @pytest.mark.skipif(SKIP_EXPECTED_ERRORS, reason=SKIP_EXPECTED_REASON)
    @pytest.mark.xfail                  # TODO: remove xfail
    def test_convert_csv_to_instance(self):
        """ensure convert_csv_to_instance works as expected"""
        debug.trace(4, "test_convert_csv_to_instance()")

        csv_data = gh.form_path(gh.dirname(__file__), "resources", "instances.csv")

        instances: List[EqCall] = THE_MODULE.convert_csv_to_instance(
            csv_data,
            "mezcla.mezcla_to_standard",
            "EqCall",
            mezcla_to_standard.EQCALL_FIELDS,
        )
        self.check_instances(instances)

    @pytest.mark.xfail                  # TODO: remove xfail
    def test_convert_py_data_to_instance(self):
        """verify convert_py_data_to_instance"""
        debug.trace(4, "test_convert_py_data_to_instance()")

        py_data = gh.form_path(gh.dirname(__file__), "resources", "instances.py-data")

        instances: List[EqCall] = THE_MODULE.convert_python_data_to_instance(
            py_data,
            "mezcla.mezcla_to_standard",
            "EqCall",
            mezcla_to_standard.EQCALL_FIELDS,
        )
        self.check_instances(instances)

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)
